from __future__ import annotations

import heapq
import keyword
import logging
import os
import re
import shutil
import string
from collections import defaultdict, deque
from copy import deepcopy
from operator import itemgetter
from pathlib import Path
from typing import (
    Any,
    DefaultDict,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)
from urllib.request import pathname2url

import eth_utils
import networkx as nx
from Crypto.Hash import BLAKE2b, keccak
from intervaltree import IntervalTree
from typing_extensions import Literal

import woke.ast.ir.type_name.mapping
import woke.ast.types as types
from woke.ast.enums import *
from woke.ast.ir.declaration.abc import DeclarationAbc
from woke.ast.ir.declaration.contract_definition import ContractDefinition
from woke.ast.ir.declaration.enum_definition import EnumDefinition
from woke.ast.ir.declaration.error_definition import ErrorDefinition
from woke.ast.ir.declaration.event_definition import EventDefinition
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.declaration.struct_definition import StructDefinition
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.meta.parameter_list import ParameterList
from woke.ast.ir.meta.source_unit import SourceUnit
from woke.ast.ir.reference_resolver import ReferenceResolver
from woke.ast.ir.statement.revert_statement import RevertStatement
from woke.ast.ir.type_name.abc import TypeNameAbc
from woke.ast.ir.type_name.array_type_name import ArrayTypeName
from woke.ast.ir.type_name.user_defined_type_name import UserDefinedTypeName
from woke.compiler import SolidityCompiler
from woke.config import WokeConfig
from woke.utils import get_package_version

from .constants import DEFAULT_IMPORTS, INIT_CONTENT, TAB_WIDTH

logger = logging.getLogger(__name__)


# TODO ensure that making the path alphanum won't create collisions
def _make_path_alphanum(source_unit_name: str) -> str:
    filtered = "".join(
        filter(lambda ch: ch.isalnum() or ch == "/" or ch == "_", source_unit_name)
    )
    return "/".join(
        f"_{segment}" if segment.startswith(tuple(string.digits)) else segment
        for segment in filtered.split("/")
    )


def _binary_search(lines: List[Tuple[bytes, int]], x: int) -> int:
    l = 0
    r = len(lines)

    while l < r:
        mid = l + (r - l) // 2
        if lines[mid][1] < x + 1:
            l = mid + 1
        else:
            r = mid

    return l - 1


def _path_to_uri(path: Path) -> str:
    if os.name == "nt":
        return "file:" + pathname2url(str(path.resolve()))
    else:
        return "file://" + pathname2url(str(path.resolve()))


def _parse_opcodes(opcodes: str) -> List[Tuple[int, str, int, Optional[int]]]:
    pc_op_map = []
    opcodes_spl = opcodes.split(" ")

    pc = 0
    ignore = False

    for i, opcode in enumerate(opcodes_spl):
        if ignore:
            ignore = False
            continue

        if not opcode.startswith("PUSH") or opcode == "PUSH0":
            pc_op_map.append((pc, opcode, 1, None))
            pc += 1
        else:
            size = int(opcode[4:]) + 1
            pc_op_map.append((pc, opcode, size, int(opcodes_spl[i + 1], 16)))
            pc += size
            ignore = True
    return pc_op_map


def _parse_source_map(
    source_map: str,
    pc_op_map: List[Tuple[int, str, int, Optional[int]]],
) -> Dict[int, Tuple[int, int, int, Optional[str]]]:
    pc_map = {}
    last_data = [-1, -1, -1, None, None]

    for i, sm_item in enumerate(source_map.split(";")):
        pc, op, size, argument = pc_op_map[i]
        source_spl = sm_item.split(":")
        for x in range(len(source_spl)):
            if source_spl[x] == "":
                continue
            if x < 3:
                last_data[x] = int(source_spl[x])
            else:
                last_data[x] = source_spl[x]

        pc_map[pc] = (
            last_data[0],
            last_data[0] + last_data[1],
            last_data[2],
            last_data[3],
        )

    return pc_map


class TypeGenerator:
    LIBRARY_PLACEHOLDER_REGEX = re.compile(r"__\$[0-9a-fA-F]{34}\$__")

    __config: WokeConfig
    __return_tx_obj: bool
    # generated types for the given source unit
    __source_unit_types: str
    # set of contracts that were already generated in the given source unit
    # used to avoid generating the same contract multiple times, eg. when multiple contracts inherit from it
    __already_generated_contracts: Set[str]
    __source_units: Dict[Path, SourceUnit]
    __interval_trees: Dict[Path, IntervalTree]
    __reference_resolver: ReferenceResolver
    __imports: SourceUnitImports
    __name_sanitizer: NameSanitizer
    __current_source_unit: str
    __pytypes_dir: Path
    __sol_to_py_lookup: Dict[str, Tuple[str, str]]
    # set of function names which should be overloaded
    __func_to_overload: Set[str]
    __contracts_index: Dict[str, Any]
    __errors_index: Dict[bytes, Dict[str, Any]]
    __events_index: Dict[bytes, Dict[str, Any]]
    __contracts_by_metadata_index: Dict[bytes, str]
    __contracts_inheritance_index: Dict[str, Tuple[str, ...]]
    __contracts_revert_index: Dict[str, Set[int]]
    __creation_code_index: List[Tuple[Tuple[Tuple[int, bytes], ...], str]]
    __line_indexes: Dict[Path, List[Tuple[bytes, int]]]
    # source unit name -> other source unit names in the cycle
    __cyclic_source_units: DefaultDict[str, Set[str]]

    def __init__(self, config: WokeConfig, return_tx_obj: bool):
        self.__config = config
        self.__return_tx_obj = return_tx_obj
        self.__source_unit_types = ""
        self.__already_generated_contracts = set()
        self.__source_units = {}
        self.__interval_trees = {}
        self.__reference_resolver = ReferenceResolver()
        self.__imports = SourceUnitImports(self)
        self.__name_sanitizer = NameSanitizer()
        self.__current_source_unit = ""
        self.__pytypes_dir = config.project_root_path / "pytypes"
        self.__sol_to_py_lookup = {}
        self.__init_sol_to_py_types()
        self.__func_to_overload = set()
        self.__contracts_index = {}
        self.__errors_index = {}
        self.__events_index = {}
        self.__contracts_by_metadata_index = {}
        self.__contracts_inheritance_index = {}
        self.__contracts_revert_index = {}
        self.__creation_code_index = []
        self.__line_indexes = {}
        self.__cyclic_source_units = defaultdict(set)

        # built-in Error(str) and Panic(uint256) errors
        error_abi = {
            "name": "Error",
            "type": "error",
            "inputs": [{"internalType": "string", "name": "message", "type": "string"}],
        }
        panic_abi = {
            "name": "Panic",
            "type": "error",
            "inputs": [{"internalType": "uint256", "name": "code", "type": "uint256"}],
        }

        for item in [error_abi, panic_abi]:
            selector = eth_utils.abi.function_abi_to_4byte_selector(item)
            self.__errors_index[selector] = {}
            self.__errors_index[selector][""] = (
                "woke.development.transactions",
                (item["name"],),
            )

    def __setup_line_indexes(self, file: Path) -> None:
        content = self.__source_units[file].file_source
        tmp_lines = re.split(b"(\r?\n)", content)
        lines: List[bytes] = []
        for line in tmp_lines:
            if line in {b"\r\n", b"\n"}:
                lines[-1] += line
            else:
                lines.append(line)

        # UTF-8 encoded lines with prefix length
        encoded_lines: List[Tuple[bytes, int]] = []
        prefix_sum = 0
        for line in lines:
            encoded_lines.append((line, prefix_sum))
            prefix_sum += len(line)
        self.__line_indexes[file] = encoded_lines

    def __get_line_pos_from_byte_offset(
        self, file: Path, byte_offset: int
    ) -> Tuple[int, int]:
        if file not in self.__line_indexes:
            self.__setup_line_indexes(file)

        encoded_lines = self.__line_indexes[file]
        line_num = _binary_search(encoded_lines, byte_offset)
        line_data, prefix_sum = encoded_lines[line_num]
        line_offset = byte_offset - prefix_sum
        return line_num, line_offset

    # TODO do some prettier init :)
    def __init_sol_to_py_types(self):
        self.__sol_to_py_lookup[types.Address.__name__] = (
            "Union[Account, Address]",
            "Address",
        )
        self.__sol_to_py_lookup[types.String.__name__] = ("str", "str")
        self.__sol_to_py_lookup[types.Bool.__name__] = ("bool", "bool")
        self.__sol_to_py_lookup[types.Bytes.__name__] = (
            "Union[bytearray, bytes]",
            "bytearray",
        )
        self.__sol_to_py_lookup[types.Function.__name__] = ("Callable", "Callable")

    @property
    def current_source_unit(self) -> str:
        return self.__current_source_unit

    @property
    def cyclic_source_units(self) -> Set[str]:
        return self.__cyclic_source_units[self.__current_source_unit]

    def add_str_to_types(
        self, num_of_indentation: int, string: str, num_of_newlines: int
    ):
        self.__source_unit_types += (
            num_of_indentation * TAB_WIDTH * " " + string + num_of_newlines * "\n"
        )

    def get_name(self, declaration: DeclarationAbc) -> str:
        return self.__name_sanitizer.sanitize_name(declaration)

    def generate_deploy_func(
        self, contract: ContractDefinition, libraries: Dict[bytes, Tuple[str, str]]
    ):
        def generate_docstring():
            if source_code_link is not None or len(param_names) > 0:
                self.add_str_to_types(2, '"""', 1)
                if source_code_link is not None:
                    self.add_str_to_types(
                        2, source_code_link, 1 if len(param_names) == 0 else 2
                    )
                if len(param_names) > 0:
                    self.add_str_to_types(2, "Args:", 1)
                    for param_name, param_type in param_names:
                        self.add_str_to_types(3, f"{param_name}: {param_type}", 1)
                self.add_str_to_types(2, '"""', 1)

        param_names = []
        params = []
        for fn in contract.functions:
            if fn.name == "constructor":
                param_names, params = self.generate_func_params(fn)
                break
        params_str = "".join(param + ", " for param in params)

        libraries_str = "".join(
            f", {l[0]}: Optional[Union[{l[1]}, Address]] = None"
            for l in libraries.values()
        )

        contract_name = self.get_name(contract)
        source_code_link = None
        for fn in contract.functions:
            if fn.kind == FunctionKind.CONSTRUCTOR:
                line, _ = self.__get_line_pos_from_byte_offset(
                    fn.file, fn.byte_location[0]
                )
                source_code_link = f"[Source code]({_path_to_uri(fn.file)}#{line + 1})"
                break

        # generate @overload stubs
        self.add_str_to_types(1, "@overload", 1)
        self.add_str_to_types(1, "@classmethod", 1)
        self.add_str_to_types(
            1,
            f"""def deploy(cls, {params_str}*, from_: Optional[Union[Account, Address, str]] = None, value: Union[int, str] = 0, gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None, return_tx: Literal[False]{'' if self.__return_tx_obj else ' = False'}{libraries_str}, request_type: Literal["call"], chain: Optional[Chain] = None, gas_price: Optional[Union[int, str]] = None, max_fee_per_gas: Optional[Union[int, str]] = None, max_priority_fee_per_gas: Optional[Union[int, str]] = None, access_list: Optional[Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]] = None, type: Optional[int] = None, block: Optional[Union[int, Literal["latest"], Literal["pending"], Literal["earliest"], Literal["safe"], Literal["finalized"]]] = None, confirmations: Optional[int] = None) -> bytearray:""",
            1,
        )
        generate_docstring()
        self.add_str_to_types(2, "...", 2)

        self.add_str_to_types(1, "@overload", 1)
        self.add_str_to_types(1, "@classmethod", 1)
        self.add_str_to_types(
            1,
            f"""def deploy(cls, {params_str}*, from_: Optional[Union[Account, Address, str]] = None, value: Union[int, str] = 0, gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None, return_tx: Literal[False]{'' if self.__return_tx_obj else ' = False'}{libraries_str}, request_type: Literal["tx"] = "tx", chain: Optional[Chain] = None, gas_price: Optional[Union[int, str]] = None, max_fee_per_gas: Optional[Union[int, str]] = None, max_priority_fee_per_gas: Optional[Union[int, str]] = None, access_list: Optional[Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]] = None, type: Optional[int] = None, block: Optional[Union[int, Literal["latest"], Literal["pending"], Literal["earliest"], Literal["safe"], Literal["finalized"]]] = None, confirmations: Optional[int] = None) -> {contract_name}:""",
            1,
        )
        generate_docstring()
        self.add_str_to_types(2, "...", 2)

        self.add_str_to_types(1, "@overload", 1)
        self.add_str_to_types(1, "@classmethod", 1)
        self.add_str_to_types(
            1,
            f"""def deploy(cls, {params_str}*, from_: Optional[Union[Account, Address, str]] = None, value: Union[int, str] = 0, gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None, return_tx: Literal[False]{'' if self.__return_tx_obj else ' = False'}{libraries_str}, request_type: Literal["estimate"], chain: Optional[Chain] = None, gas_price: Optional[Union[int, str]] = None, max_fee_per_gas: Optional[Union[int, str]] = None, max_priority_fee_per_gas: Optional[Union[int, str]] = None, access_list: Optional[Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]] = None, type: Optional[int] = None, block: Optional[Union[int, Literal["latest"], Literal["pending"], Literal["earliest"], Literal["safe"], Literal["finalized"]]] = None, confirmations: Optional[int] = None) -> int:""",
            1,
        )
        generate_docstring()
        self.add_str_to_types(2, "...", 2)

        self.add_str_to_types(1, "@overload", 1)
        self.add_str_to_types(1, "@classmethod", 1)
        self.add_str_to_types(
            1,
            f"""def deploy(cls, {params_str}*, from_: Optional[Union[Account, Address, str]] = None, value: Union[int, str] = 0, gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None, return_tx: Literal[False]{'' if self.__return_tx_obj else ' = False'}{libraries_str}, request_type: Literal["access_list"], chain: Optional[Chain] = None, gas_price: Optional[Union[int, str]] = None, max_fee_per_gas: Optional[Union[int, str]] = None, max_priority_fee_per_gas: Optional[Union[int, str]] = None, access_list: Optional[Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]] = None, type: Optional[int] = None, block: Optional[Union[int, Literal["latest"], Literal["pending"], Literal["earliest"], Literal["safe"], Literal["finalized"]]] = None, confirmations: Optional[int] = None) -> Tuple[Dict[Address, List[int]], int]:""",
            1,
        )
        generate_docstring()
        self.add_str_to_types(2, "...", 2)

        self.add_str_to_types(1, "@overload", 1)
        self.add_str_to_types(1, "@classmethod", 1)
        self.add_str_to_types(
            1,
            f"""def deploy(cls, {params_str}*, from_: Optional[Union[Account, Address, str]] = None, value: Union[int, str] = 0, gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None, return_tx: Literal[True]{' = True' if self.__return_tx_obj else ''}{libraries_str}, request_type: Literal["tx"] = "tx", chain: Optional[Chain] = None, gas_price: Optional[Union[int, str]] = None, max_fee_per_gas: Optional[Union[int, str]] = None, max_priority_fee_per_gas: Optional[Union[int, str]] = None, access_list: Optional[Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]] = None, type: Optional[int] = None, block: Optional[Union[int, Literal["latest"], Literal["pending"], Literal["earliest"], Literal["safe"], Literal["finalized"]]] = None, confirmations: Optional[int] = None) -> TransactionAbc[{contract_name}]:""",
            1,
        )
        generate_docstring()
        self.add_str_to_types(2, "...", 2)

        self.add_str_to_types(1, "@classmethod", 1)
        self.add_str_to_types(
            1,
            f'def deploy(cls, {params_str}*, from_: Optional[Union[Account, Address, str]] = None, value: Union[int, str] = 0, gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None, return_tx: bool = {self.__return_tx_obj}{libraries_str}, request_type: RequestType = "tx", chain: Optional[Chain] = None, gas_price: Optional[Union[int, str]] = None, max_fee_per_gas: Optional[Union[int, str]] = None, max_priority_fee_per_gas: Optional[Union[int, str]] = None, access_list: Optional[Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]] = None, type: Optional[int] = None, block: Optional[Union[int, Literal["latest"], Literal["pending"], Literal["earliest"], Literal["safe"], Literal["finalized"]]] = None, confirmations: Optional[int] = None) -> Union[bytearray, {contract_name}, int, Tuple[Dict[Address, List[int]], int], TransactionAbc[{contract_name}]]:',
            1,
        )

        generate_docstring()

        if contract.kind in {ContractKind.CONTRACT, ContractKind.LIBRARY}:
            if not contract.abstract:
                libs_arg = (
                    "{"
                    + ", ".join(
                        f"{lib_id}: ({l[0]}, '{l[1]}')"
                        for lib_id, l in libraries.items()
                    )
                    + "}"
                )
                self.add_str_to_types(
                    2,
                    f"return cls._deploy(request_type, [{', '.join(map(itemgetter(0), param_names))}], return_tx, {contract_name}, from_, value, gas_limit, {libs_arg}, chain, gas_price, max_fee_per_gas, max_priority_fee_per_gas, access_list, type, block, confirmations)",
                    1,
                )
            else:
                self.add_str_to_types(
                    2, 'raise Exception("Cannot deploy abstract contract")', 1
                )
        else:
            self.add_str_to_types(2, 'raise Exception("Cannot deploy interface")', 1)

    def generate_creation_code_func(
        self, contract: ContractDefinition, libraries: Dict[bytes, Tuple[str, str]]
    ):
        libraries_arg = "".join(
            f", {l[0]}: Union[{l[1]}, Address]" for l in libraries.values()
        )

        self.add_str_to_types(1, "@classmethod", 1)
        self.add_str_to_types(
            1,
            f"def get_creation_code(cls{libraries_arg}) -> bytes:",
            1,
        )

        if contract.kind in {ContractKind.CONTRACT, ContractKind.LIBRARY}:
            if not contract.abstract:
                libs_arg = (
                    "{"
                    + ", ".join(
                        f"{lib_id}: ({l[0]}, '{l[1]}')"
                        for lib_id, l in libraries.items()
                    )
                    + "}"
                )
                self.add_str_to_types(
                    2,
                    f"return cls._get_creation_code({libs_arg})",
                    1,
                )
            else:
                self.add_str_to_types(
                    2,
                    'raise Exception("Cannot get creation code of an abstract contract")',
                    1,
                )
        else:
            self.add_str_to_types(
                2, 'raise Exception("Cannot get creation code of an interface")', 1
            )

    def generate_contract_template(
        self, contract: ContractDefinition, base_names: str
    ) -> Dict[bytes, Any]:
        if contract.kind == ContractKind.LIBRARY:
            self.add_str_to_types(
                0, "class " + self.get_name(contract) + "(Library):", 1
            )
        else:
            self.add_str_to_types(
                0, "class " + self.get_name(contract) + "(" + base_names + "):", 1
            )
        line, _ = self.__get_line_pos_from_byte_offset(
            contract.file, contract.byte_location[0]
        )
        self.add_str_to_types(1, '"""', 1)
        self.add_str_to_types(
            1, f"[Source code]({_path_to_uri(contract.file)}#{line + 1})", 1
        )
        self.add_str_to_types(1, '"""', 1)

        compilation_info = contract.compilation_info
        assert compilation_info is not None
        assert compilation_info.abi is not None
        assert compilation_info.evm is not None
        assert compilation_info.evm.bytecode is not None
        assert compilation_info.evm.bytecode.object is not None
        assert compilation_info.evm.deployed_bytecode is not None
        assert compilation_info.evm.deployed_bytecode.object is not None
        assert compilation_info.evm.deployed_bytecode.opcodes is not None
        assert compilation_info.evm.deployed_bytecode.source_map is not None

        fqn = f"{contract.parent.source_unit_name}:{contract.name}"
        parsed_opcodes = _parse_opcodes(compilation_info.evm.deployed_bytecode.opcodes)
        pc_map = _parse_source_map(
            compilation_info.evm.deployed_bytecode.source_map, parsed_opcodes
        )

        for pc, op, size, argument in parsed_opcodes:
            if op == "REVERT" and pc in pc_map:
                start, end, file_id, _ = pc_map[pc]
                if file_id == -1:
                    continue
                try:
                    path = self.__reference_resolver.resolve_source_file_id(
                        file_id, contract.cu_hash
                    )
                except KeyError:
                    continue

                intervals = self.__interval_trees[path].envelop(start, end)
                nodes: List = sorted(
                    [interval.data for interval in intervals],
                    key=lambda n: n.ast_tree_depth,
                )

                if len(nodes) > 0:
                    node = nodes[0]
                    if isinstance(node, FunctionCall) and isinstance(
                        node.parent, RevertStatement
                    ):
                        func_called = node.function_called
                        assert isinstance(func_called, ErrorDefinition)

                        if fqn not in self.__contracts_revert_index:
                            self.__contracts_revert_index[fqn] = set()
                        self.__contracts_revert_index[fqn].add(pc)

        if len(compilation_info.evm.deployed_bytecode.object) > 0:
            metadata = bytes.fromhex(
                compilation_info.evm.deployed_bytecode.object[-106:]
            )
            assert len(metadata) == 53
            assert metadata not in self.__contracts_by_metadata_index
            self.__contracts_by_metadata_index[metadata] = fqn

        assert (
            fqn not in self.__contracts_inheritance_index
        ), f"Generating contract {fqn} twice"
        self.__contracts_inheritance_index[fqn] = tuple(
            f"{base.parent.source_unit_name}:{base.name}"
            for base in contract.linearized_base_contracts
        )

        abi_by_selector: Dict[
            Union[
                bytes, Literal["constructor"], Literal["fallback"], Literal["receive"]
            ],
            Dict,
        ] = {}

        module_name = "pytypes." + _make_path_alphanum(
            contract.parent.source_unit_name[:-3]
        ).replace("/", ".")

        events_abi = {}

        for item in compilation_info.abi:
            if item["type"] == "function":
                if contract.kind == ContractKind.LIBRARY:
                    item_copy = deepcopy(item)
                    for arg in item_copy["inputs"]:
                        if arg["internalType"].startswith("contract "):
                            arg["internalType"] = arg["internalType"][9:]
                        elif arg["internalType"].startswith("struct "):
                            arg["internalType"] = arg["internalType"][7:]
                        elif arg["internalType"].startswith("enum "):
                            arg["internalType"] = arg["internalType"][5:]

                    selector = keccak.new(
                        data=f"{item['name']}({','.join(arg['internalType'] for arg in item_copy['inputs'])})".encode(
                            "utf-8"
                        ),
                        digest_bits=256,
                    ).digest()[:4]
                else:
                    selector = eth_utils.abi.function_abi_to_4byte_selector(item)
                abi_by_selector[selector] = item
            elif item["type"] == "error":
                selector = eth_utils.abi.function_abi_to_4byte_selector(item)

                if selector not in self.__errors_index:
                    self.__errors_index[selector] = {}

                # find where the error is declared
                error_decl = None
                for error in contract.used_errors:
                    if error.name == item["name"]:
                        error_decl = error
                        break
                assert error_decl is not None

                if isinstance(error_decl.parent, ContractDefinition):
                    # error is declared in a contract
                    error_module_name = "pytypes." + _make_path_alphanum(
                        error_decl.parent.parent.source_unit_name[:-3]
                    ).replace("/", ".")
                    self.__errors_index[selector][fqn] = (
                        error_module_name,
                        (self.get_name(error_decl.parent), self.get_name(error_decl)),
                    )
                elif isinstance(error_decl.parent, SourceUnit):
                    error_module_name = "pytypes." + _make_path_alphanum(
                        error_decl.parent.source_unit_name[:-3]
                    ).replace("/", ".")
                    self.__errors_index[selector][fqn] = (
                        error_module_name,
                        (self.get_name(error_decl),),
                    )
                else:
                    raise Exception("Unknown error parent")
            elif item["type"] == "event":
                selector = eth_utils.abi.event_abi_to_log_topic(item)
                events_abi[selector] = item

                event_decl = None
                for c in contract.linearized_base_contracts:
                    for event in c.events:
                        if event.event_selector == selector:
                            event_decl = event
                            break
                    if event_decl is not None:
                        break
                if event_decl is None:
                    continue

                if selector not in self.__events_index:
                    self.__events_index[selector] = {}
                event_module_name = "pytypes." + _make_path_alphanum(
                    event_decl.parent.parent.source_unit_name[:-3]
                ).replace("/", ".")
                self.__events_index[selector][fqn] = (
                    event_module_name,
                    (self.get_name(event_decl.parent), self.get_name(event_decl)),
                )
            elif item["type"] in {"constructor", "fallback", "receive"}:
                abi_by_selector[item["type"]] = item
            else:
                raise Exception(f"Unexpected ABI item type: {item['type']}")
        self.add_str_to_types(1, f"_abi = {abi_by_selector}", 1)
        self.add_str_to_types(
            1, f'_creation_code = "{compilation_info.evm.bytecode.object}"', 2
        )

        if contract.kind == ContractKind.LIBRARY:
            lib_id = keccak.new(data=fqn.encode("utf-8"), digest_bits=256).digest()[:17]
            self.add_str_to_types(1, f"_library_id = {lib_id}", 2)

        # find all needed libraries
        lib_ids: Set[bytes] = set()
        bytecode = compilation_info.evm.bytecode.object

        if len(bytecode) > 0:
            bytecode_segments: List[Tuple[int, bytes]] = []
            start = 0

            # TODO test segments work in this case
            for match in self.__class__.LIBRARY_PLACEHOLDER_REGEX.finditer(bytecode):
                s = match.start()
                e = match.end()
                segment = bytes.fromhex(bytecode[start:s])
                h = BLAKE2b.new(data=segment, digest_bits=256).digest()
                bytecode_segments.append((len(segment), h))
                start = e
                lib_id = bytes.fromhex(bytecode[s + 3 : e - 3])
                lib_ids.add(lib_id)

            fqn = f"{contract.parent.source_unit_name}:{contract.name}"

            segment = bytes.fromhex(bytecode[start:])
            h = BLAKE2b.new(data=segment, digest_bits=256).digest()
            bytecode_segments.append((len(segment), h))

            self.__creation_code_index.append((tuple(bytecode_segments), fqn))

        libraries: Dict[bytes, Tuple[str, str]] = {}
        source_units_queue = deque([contract.parent])

        while len(source_units_queue) > 0 and len(lib_ids) > 0:
            source_unit = source_units_queue.popleft()
            for c in source_unit.contracts:
                if c.kind == ContractKind.LIBRARY:
                    fqn = f"{c.parent.source_unit_name}:{c.name}"
                    lib_id = keccak.new(
                        data=fqn.encode("utf-8"), digest_bits=256
                    ).digest()[:17]

                    if lib_id in lib_ids:
                        lib_ids.remove(lib_id)
                        self.__imports.generate_contract_import(c)
                        libraries[lib_id] = (
                            c.name[0].lower() + c.name[1:],
                            self.get_name(c),
                        )

            source_units_queue.extend(imp.source_unit for imp in source_unit.imports)

        assert len(lib_ids) == 0, "Not all libraries were found"

        self.generate_deploy_func(contract, libraries)
        self.add_str_to_types(0, "", 1)
        self.generate_creation_code_func(contract, libraries)
        self.add_str_to_types(0, "", 1)

        return events_abi

    def generate_types_struct(
        self, structs: Iterable[StructDefinition], indent: int
    ) -> None:
        for struct in structs:
            members: List[Tuple[str, str, str, str]] = []
            for member in struct.members:
                member_name = self.get_name(member)
                member_type = self.parse_type_and_import(member.type, True)
                member_type_desc = member.type_string
                members.append(
                    (member_name, member_type, member_type_desc, member.name)
                )

            self.add_str_to_types(indent, "@dataclasses.dataclass", 1)
            self.add_str_to_types(indent, f"class {self.get_name(struct)}:", 1)
            self.add_str_to_types(indent + 1, '"""', 1)
            line, _ = self.__get_line_pos_from_byte_offset(
                struct.file, struct.byte_location[0]
            )
            self.add_str_to_types(
                indent + 1, f"[Source code]({_path_to_uri(struct.file)}#{line + 1})", 2
            )
            self.add_str_to_types(indent + 1, "Attributes:", 1)
            for member_name, member_type, member_type_desc, _ in members:
                self.add_str_to_types(
                    indent + 2, f"{member_name} ({member_type}): {member_type_desc}", 1
                )
            self.add_str_to_types(indent + 1, '"""', 1)
            self.add_str_to_types(indent + 1, f"original_name = '{struct.name}'", 2)

            for member_name, member_type, _, original_name in members:
                if member_name == original_name:
                    self.add_str_to_types(
                        indent + 1, f"{member_name}: {member_type}", 1
                    )
                else:
                    self.add_str_to_types(
                        indent + 1,
                        f'{member_name}: {member_type} = dataclasses.field(metadata={{"original_name": "{original_name}"}})',
                        1,
                    )
            self.add_str_to_types(0, "", 2)

    def generate_types_enum(self, enums: Iterable[EnumDefinition], indent: int) -> None:
        self.__imports.add_python_import("from enum import IntEnum")
        for enum in enums:
            self.add_str_to_types(indent, f"class {self.get_name(enum)}(IntEnum):", 1)
            line, _ = self.__get_line_pos_from_byte_offset(
                enum.file, enum.byte_location[0]
            )
            self.add_str_to_types(indent + 1, '"""', 1)
            self.add_str_to_types(
                indent + 1, f"[Source code]({_path_to_uri(enum.file)}#{line + 1})", 2
            )
            self.add_str_to_types(indent + 1, '"""', 1)
            num = 0
            for member in enum.values:
                self.add_str_to_types(
                    indent + 1, self.get_name(member) + " = " + str(num), 1
                )
                num += 1
            self.add_str_to_types(0, "", 2)

    def generate_types_error(
        self,
        errors: Iterable[ErrorDefinition],
        indent: int,
    ) -> None:
        for error in errors:
            # cannot generate pytypes for unused errors
            if len(error.used_in) == 0:
                continue

            used_in = error.used_in[0]
            assert used_in.compilation_info is not None
            assert used_in.compilation_info.abi is not None

            error_abi = None

            for item in used_in.compilation_info.abi:
                if item["type"] == "error" and item["name"] == error.name:
                    selector = eth_utils.abi.function_abi_to_4byte_selector(item)
                    if selector == error.error_selector:
                        error_abi = item
                        break

            assert error_abi is not None

            parameters: List[Tuple[str, str, str, str]] = []
            unnamed_params_index = 1
            for parameter in error.parameters.parameters:
                if parameter.name:
                    parameter_name = self.get_name(parameter)
                else:
                    parameter_name = f"param{unnamed_params_index}"
                    unnamed_params_index += 1
                parameter_type = self.parse_type_and_import(parameter.type, True)
                parameter_type_desc = parameter.type_string
                parameters.append(
                    (
                        parameter_name,
                        parameter_type,
                        parameter_type_desc,
                        parameter.name,
                    )
                )

            self.add_str_to_types(indent, "@dataclasses.dataclass", 1)
            self.add_str_to_types(
                indent,
                f"class {self.get_name(error)}(TransactionRevertedError):",
                1,
            )

            line, _ = self.__get_line_pos_from_byte_offset(
                error.file, error.byte_location[0]
            )
            self.add_str_to_types(indent + 1, '"""', 1)
            self.add_str_to_types(
                indent + 1,
                f"[Source code]({_path_to_uri(error.file)}#{line + 1})",
                1 if len(parameters) == 0 else 2,
            )

            if len(parameters) > 0:
                self.add_str_to_types(indent + 1, "Attributes:", 1)
                for param_name, param_type, param_type_desc, _ in parameters:
                    self.add_str_to_types(
                        indent + 2, f"{param_name} ({param_type}): {param_type_desc}", 1
                    )

            self.add_str_to_types(indent + 1, '"""', 1)
            self.add_str_to_types(indent + 1, f"_abi = {error_abi}", 1)
            self.add_str_to_types(indent + 1, f"original_name = '{error.name}'", 1)
            self.add_str_to_types(indent + 1, f"selector = {error.error_selector}", 2)
            for param_name, param_type, _, original_name in parameters:
                if param_name == original_name:
                    self.add_str_to_types(indent + 1, f"{param_name}: {param_type}", 1)
                else:
                    self.add_str_to_types(
                        indent + 1,
                        f'{param_name}: {param_type} = dataclasses.field(metadata={{"original_name": "{original_name}"}})',
                        1,
                    )
            self.add_str_to_types(0, "", 2)

    def generate_types_event(
        self,
        events: Iterable[EventDefinition],
        indent: int,
        events_abi: Dict[bytes, Any],
    ) -> None:
        for event in events:
            parameters: List[Tuple[str, str, str, str]] = []
            unnamed_params_index = 1
            for parameter in event.parameters.parameters:
                if parameter.name:
                    parameter_name = self.get_name(parameter)
                else:
                    parameter_name = f"param{unnamed_params_index}"
                    unnamed_params_index += 1

                if parameter.indexed and isinstance(
                    parameter.type,
                    (types.Array, types.Struct, types.Bytes, types.String),
                ):
                    parameter_name += "_hash"
                    parameter_type = "bytes"
                else:
                    parameter_type = self.parse_type_and_import(parameter.type, True)
                if parameter.indexed:
                    parameter_type_desc = "indexed " + parameter.type_string
                else:
                    parameter_type_desc = parameter.type_string
                parameters.append(
                    (
                        parameter_name,
                        parameter_type,
                        parameter_type_desc,
                        parameter.name,
                    )
                )

            self.add_str_to_types(indent, "@dataclasses.dataclass", 1)
            self.add_str_to_types(indent, f"class {self.get_name(event)}:", 1)

            line, _ = self.__get_line_pos_from_byte_offset(
                event.file, event.byte_location[0]
            )
            self.add_str_to_types(indent + 1, '"""', 1)
            self.add_str_to_types(
                indent + 1,
                f"[Source code]({_path_to_uri(event.file)}#{line + 1})",
                1 if len(parameters) == 0 else 2,
            )

            if len(parameters) > 0:
                self.add_str_to_types(indent + 1, "Attributes:", 1)
                for param_name, param_type, param_type_desc, _ in parameters:
                    self.add_str_to_types(
                        indent + 2, f"{param_name} ({param_type}): {param_type_desc}", 1
                    )

            self.add_str_to_types(indent + 1, '"""', 1)
            self.add_str_to_types(
                indent + 1, f"_abi = {events_abi[event.event_selector]}", 1
            )
            self.add_str_to_types(indent + 1, f"original_name = '{event.name}'", 1)
            self.add_str_to_types(indent + 1, f"selector = {event.event_selector}", 2)
            for param_name, param_type, _, original_name in parameters:
                if param_name == original_name:
                    self.add_str_to_types(indent + 1, f"{param_name}: {param_type}", 1)
                else:
                    self.add_str_to_types(
                        indent + 1,
                        f'{param_name}: {param_type} = dataclasses.field(metadata={{"original_name": "{original_name}"}})',
                        1,
                    )
            self.add_str_to_types(0, "", 2)

    # parses the expr to string
    # optionaly generates an import
    def parse_type_and_import(self, expr: types.TypeAbc, return_types: bool) -> str:
        if return_types:
            types_index = 1
        else:
            types_index = 0

        if isinstance(expr, types.Struct):
            parent = expr.ir_node.parent
            if isinstance(parent, ContractDefinition):
                self.__imports.generate_contract_import(parent)
                return f"{self.get_name(parent)}.{self.get_name(expr.ir_node)}"
            else:
                self.__imports.generate_struct_import(expr)
                return self.get_name(expr.ir_node)
        elif isinstance(expr, types.Enum):
            parent = expr.ir_node.parent
            if isinstance(parent, ContractDefinition):
                self.__imports.generate_contract_import(parent)
                return f"{self.get_name(parent)}.{self.get_name(expr.ir_node)}"
            else:
                self.__imports.generate_enum_import(expr)
                return self.get_name(expr.ir_node)
        elif isinstance(expr, types.UserDefinedValueType):
            return self.parse_type_and_import(
                expr.ir_node.underlying_type.type, return_types
            )
        elif isinstance(expr, types.Array):
            if expr.length is None:
                return (
                    f"List[{self.parse_type_and_import(expr.base_type, return_types)}]"
                )
            elif expr.length <= 32:
                return f"List{expr.length}[{self.parse_type_and_import(expr.base_type, return_types)}]"
            else:
                return f"Annotated[List[{self.parse_type_and_import(expr.base_type, return_types)}], Length({expr.length})]"
        elif isinstance(expr, types.Contract):
            self.__imports.generate_contract_import(expr.ir_node)
            return self.get_name(expr.ir_node)
        elif isinstance(expr, types.FixedBytes):
            return f"bytes{expr.bytes_count}"
        elif isinstance(expr, types.Int):
            return f"int{expr.bits_count}"
        elif isinstance(expr, types.UInt):
            return f"uint{expr.bits_count}"
        elif isinstance(expr, types.Mapping):
            return f"Dict[{self.parse_type_and_import(expr.key_type, return_types)}, {self.parse_type_and_import(expr.value_type, return_types)}]"
        else:
            return self.__sol_to_py_lookup[expr.__class__.__name__][types_index]

    def generate_func_params(
        self, fn: FunctionDefinition
    ) -> Tuple[List[Tuple[str, str]], List[str]]:
        params = []
        param_names = []
        unnamed_params_identifier: int = 1
        generated_names = {
            self.get_name(par) for par in fn.parameters.parameters if par.name != ""
        }

        for param in fn.parameters.parameters:
            if param.name == "":
                param_name: str = "arg" + str(unnamed_params_identifier)
                unnamed_params_identifier += 1
                while param_name in generated_names:
                    param_name += "_"
            else:
                param_name = self.get_name(param)
            param_names.append((param_name, param.type_string))
            params.append(
                f"{param_name}: {self.parse_type_and_import(param.type, False)}"
            )
        return param_names, params

    def generate_func_returns(self, fn: FunctionDefinition) -> List[Tuple[str, str]]:
        self.__imports.clear_tmp_type_checking_imports()
        return [
            (self.parse_type_and_import(par.type, True), par.type_string)
            for par in fn.return_parameters.parameters
        ]

    def is_compound_type(self, var_type: types.TypeAbc):
        name = var_type.__class__.__name__
        return name == "Array" or name == "Mapping"

    def generate_getter_for_state_var(self, decl: VariableDeclaration):
        def get_struct_return_list(
            struct_type_name: UserDefinedTypeName,
        ) -> List[Tuple[str, str]]:
            struct = struct_type_name.type
            assert isinstance(struct, types.Struct)
            node = struct.ir_node
            non_excluded: List[Tuple[str, str]] = []
            for member in node.members:
                if not isinstance(member.type, types.Mapping) and not isinstance(
                    member.type, types.Array
                ):
                    non_excluded.append(
                        (
                            self.parse_type_and_import(member.type, True),
                            member.type_string,
                        )
                    )
            if len(node.members) == len(non_excluded):
                # nothing was excluded -> the whole struct will be used -> add the struct to imports
                parent = node.parent
                if isinstance(parent, ContractDefinition):
                    self.__imports.generate_contract_import(parent)
                    return [
                        (
                            f"{self.get_name(parent)}.{self.get_name(struct.ir_node)}",
                            struct_type_name.type_string,
                        )
                    ]
                else:
                    self.__imports.generate_struct_import(struct)
                    return [
                        (self.get_name(struct.ir_node), struct_type_name.type_string)
                    ]
            else:
                return non_excluded

        returns: List[Tuple[str, str]] = []
        param_names: List[Tuple[str, str]] = []
        # if the type is compound we need to use the type as an index, for primitive types we use the
        # the type only for the return
        # TODO reorder the elif chain such that the most common types are on the top
        def generate_getter_helper(
            var_type_name: TypeNameAbc, use_parse: bool, depth: int
        ) -> List[str]:
            nonlocal returns
            nonlocal param_names
            parsed = []
            var_type = var_type_name.type
            if isinstance(var_type, types.Struct):
                if depth == 0:
                    pass
                else:
                    parent = var_type.ir_node.parent
                    if isinstance(parent, ContractDefinition):
                        self.__imports.generate_contract_import(parent)
                        parsed.append(
                            f"{self.get_name(parent)}.{self.get_name(var_type.ir_node)}"
                        )
                    else:
                        self.__imports.generate_struct_import(var_type)
                        parsed.append(self.get_name(var_type.ir_node))
                assert isinstance(var_type_name, UserDefinedTypeName)
                returns = get_struct_return_list(var_type_name)
            elif isinstance(var_type, types.Enum):
                parent = var_type.ir_node.parent
                if isinstance(parent, ContractDefinition):
                    self.__imports.generate_contract_import(parent)
                    parsed.append(
                        f"{self.get_name(parent)}.{self.get_name(var_type.ir_node)}"
                    )
                    returns = [
                        (
                            f"{self.get_name(parent)}.{self.get_name(var_type.ir_node)}",
                            var_type_name.type_string,
                        )
                    ]

                else:
                    self.__imports.generate_enum_import(var_type)
                    parsed.append(self.get_name(var_type.ir_node))
                    returns = [
                        (self.get_name(var_type.ir_node), var_type_name.type_string)
                    ]
            elif isinstance(var_type, types.UserDefinedValueType):
                underlying_type = var_type.ir_node.underlying_type.type
                if isinstance(underlying_type, types.FixedBytes):
                    parsed.append(f"bytes{underlying_type.bytes_count}")
                    returns = [
                        (
                            f"bytes{underlying_type.bytes_count}",
                            var_type_name.type_string,
                        )
                    ]
                elif isinstance(underlying_type, types.Int):
                    parsed.append(f"int{underlying_type.bits_count}")
                    returns = [
                        (f"int{underlying_type.bits_count}", var_type_name.type_string)
                    ]
                elif isinstance(underlying_type, types.UInt):
                    parsed.append(f"uint{underlying_type.bits_count}")
                    returns = [
                        (f"uint{underlying_type.bits_count}", var_type_name.type_string)
                    ]
                else:
                    parsed.append(
                        self.__sol_to_py_lookup[underlying_type.__class__.__name__][0]
                    )
                    returns = [
                        (
                            self.__sol_to_py_lookup[underlying_type.__class__.__name__][
                                1
                            ],
                            var_type_name.type_string,
                        )
                    ]
            elif isinstance(var_type, types.Array):
                use_parse = True
                param_names.append(("index" + str(depth), "uint256"))
                parsed.append(f"index{depth}: int")
                assert isinstance(var_type_name, ArrayTypeName)
                if self.is_compound_type(var_type.base_type):
                    parsed.extend(
                        generate_getter_helper(var_type_name.base_type, True, depth + 1)
                    )
                else:
                    # ignores the parsed return, only called for the side-effect of changing the returns var to value_type
                    _ = generate_getter_helper(
                        var_type_name.base_type, False, depth + 1
                    )
            elif isinstance(var_type, types.Mapping):
                # parse key
                use_parse = True
                assert isinstance(var_type_name, woke.ast.ir.type_name.mapping.Mapping)
                param_names.append(
                    ("key" + str(depth), var_type_name.key_type.type_string)
                )
                key_type = generate_getter_helper(
                    var_type_name.key_type, True, depth + 1
                )
                assert len(key_type) == 1
                parsed.append(f"key{depth}: {key_type[0]}")
                if self.is_compound_type(var_type.value_type):
                    parsed.extend(
                        generate_getter_helper(
                            var_type_name.value_type, True, depth + 1
                        )
                    )
                else:
                    # ignores the parsed return, only called for the side-effect of changing the returns var to value_type
                    _ = generate_getter_helper(
                        var_type_name.value_type, True, depth + 1
                    )
            elif isinstance(var_type, types.Contract):
                self.__imports.generate_contract_import(var_type.ir_node)
                parsed.append(self.get_name(var_type.ir_node))
                returns = [(self.get_name(var_type.ir_node), var_type_name.type_string)]
            elif isinstance(var_type, types.FixedBytes):
                parsed.append(f"bytes{var_type.bytes_count}")
                returns = [(f"bytes{var_type.bytes_count}", var_type_name.type_string)]
            elif isinstance(var_type, types.Int):
                parsed.append(f"int{var_type.bits_count}")
                returns = [(f"int{var_type.bits_count}", var_type_name.type_string)]
            elif isinstance(var_type, types.UInt):
                parsed.append(f"uint{var_type.bits_count}")
                returns = [(f"uint{var_type.bits_count}", var_type_name.type_string)]
            else:
                parsed.append(self.__sol_to_py_lookup[var_type.__class__.__name__][0])
                returns = [
                    (
                        self.__sol_to_py_lookup[var_type.__class__.__name__][1],
                        var_type_name.type_string,
                    )
                ]

            return parsed if use_parse else []

        self.__imports.clear_tmp_type_checking_imports()
        generated_params = generate_getter_helper(decl.type_name, False, 0)

        if len(returns) == 0:
            returns_str = "None"
        elif len(returns) == 1:
            returns_str = returns[0][0]
        else:
            returns_str = f"Tuple[{', '.join(ret[0] for ret in returns)}]"

        self.generate_type_hint_stub_func(
            decl, generated_params, returns_str, "call", True, param_names, returns
        )
        self.generate_type_hint_stub_func(
            decl, generated_params, "int", "estimate", False, param_names, returns
        )
        self.generate_type_hint_stub_func(
            decl,
            generated_params,
            "Tuple[Dict[Address, List[int]], int]",
            "access_list",
            False,
            param_names,
            returns,
        )
        self.generate_type_hint_stub_func(
            decl,
            generated_params,
            f"TransactionAbc[{returns_str}]",
            "tx",
            False,
            param_names,
            returns,
        )

        self.generate_func_implementation(
            decl,
            generated_params,
            param_names,
            returns,
        )

    # receives names of params and their type hints, returns only the types to be used for dispatch
    def get_types_from_func_params(self, params) -> str:
        # 1. split on , to separete the params
        # 2. split on : and get the last (the second) elem to get the type (each pair is in the format name: type)
        # 3. remove the last elem which is params: Optional[TxParams] = None (not used for dispatch)
        name_type = params.split(",")
        types = []
        for name_type_pair in name_type:
            types.append(name_type_pair.split(":")[-1])
        res: str = ", ".join(types[:-1])
        if res:
            # remove the first char which a redundant whitespace
            res = res[1:]
        return res

    def generate_func_implementation(
        self,
        declaration: Union[FunctionDefinition, VariableDeclaration],
        params: List[str],
        param_names: List[Tuple[str, str]],
        returns: List[Tuple[str, str]],
    ):
        is_view_or_pure: bool = isinstance(
            declaration, VariableDeclaration
        ) or declaration.state_mutability in {
            StateMutability.VIEW,
            StateMutability.PURE,
        }
        params_str = "".join(param + ", " for param in params)
        if len(returns) == 0:
            returns_str = None
        elif len(returns) == 1:
            returns_str = returns[0][0]
        else:
            returns_str = f"Tuple[{', '.join(ret[0] for ret in returns)}]"
        self.add_str_to_types(
            1,
            f"""def {self.get_name(declaration)}(self, {params_str}*, from_: Optional[Union[Account, Address, str]] = None, to: Optional[Union[Account, Address, str]] = None, value: Union[int, str] = 0, gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None, request_type: RequestType = '{'call' if is_view_or_pure else 'tx'}', gas_price: Optional[Union[int, str]] = None, max_fee_per_gas: Optional[Union[int, str]] = None, max_priority_fee_per_gas: Optional[Union[int, str]] = None, access_list: Optional[Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]] = None, type: Optional[int] = None, block: Optional[Union[int, Literal["latest"], Literal["pending"], Literal["earliest"], Literal["safe"], Literal["finalized"]]] = None, confirmations: Optional[int] = None) -> Union[{returns_str}, TransactionAbc[{returns_str}], int, Tuple[Dict[Address, List[int]], int]]:""",
            1,
        )

        line, _ = self.__get_line_pos_from_byte_offset(
            declaration.file, declaration.byte_location[0]
        )
        self.add_str_to_types(2, '"""', 1)
        self.add_str_to_types(
            2,
            f"[Source code]({_path_to_uri(declaration.file)}#{line + 1})",
            1 if len(param_names) + len(returns) == 0 else 2,
        )
        if len(param_names) > 0:
            self.add_str_to_types(2, "Args:", 1)
            for param_name, param_type in param_names:
                self.add_str_to_types(3, f"{param_name}: {param_type}", 1)
        if len(returns) == 1:
            self.add_str_to_types(2, "Returns:", 1)
            self.add_str_to_types(3, f"{returns[0][1]}", 1)
        elif len(returns) > 1:
            self.add_str_to_types(2, "Returns:", 1)
            self.add_str_to_types(3, f'({", ".join(ret[1] for ret in returns)})', 1)
        self.add_str_to_types(2, '"""', 1)

        if len(returns) == 0:
            return_types = "NoneType"
        elif len(returns) == 1:
            return_types = returns[0][0]
        else:
            return_types = f"Tuple[{', '.join(map(itemgetter(0), returns))}]"

        for cyclic_import in self.__imports.tmp_type_checking_imports:
            self.add_str_to_types(2, cyclic_import, 1)
        if self.__imports.tmp_type_checking_imports:
            self.add_str_to_types(2, "", 1)

        assert declaration.function_selector is not None
        fn_selector = declaration.function_selector.hex()
        self.add_str_to_types(
            2,
            f'return self._execute(self.chain, request_type, "{fn_selector}", [{", ".join(map(itemgetter(0), param_names))}], True if request_type == "tx" else False, {return_types}, from_, to if to is not None else str(self.address), value, gas_limit, gas_price, max_fee_per_gas, max_priority_fee_per_gas, access_list, type, block, confirmations)',
            2,
        )

    def generate_type_hint_stub_func(
        self,
        declaration: Union[FunctionDefinition, VariableDeclaration],
        params: List[str],
        returns_str: str,
        request_type: str,
        request_type_is_default: bool,
        param_names: List[Tuple[str, str]],
        returns: List[Tuple[str, str]],
    ):
        params_str = "".join(param + ", " for param in params)

        self.add_str_to_types(1, "@overload", 1)
        self.add_str_to_types(
            1,
            f"""def {self.get_name(declaration)}(self, {params_str}*, from_: Optional[Union[Account, Address, str]] = None, to: Optional[Union[Account, Address, str]] = None, value: Union[int, str] = 0, gas_limit: Optional[Union[int, Literal["max"], Literal["auto"]]] = None, request_type: Literal["{request_type}"]{' = "' + request_type + '"' if request_type_is_default else ''}, gas_price: Optional[Union[int, str]] = None, max_fee_per_gas: Optional[Union[int, str]] = None, max_priority_fee_per_gas: Optional[Union[int, str]] = None, access_list: Optional[Union[Dict[Union[Account, Address, str], List[int]], Literal["auto"]]] = None, type: Optional[int] = None, block: Optional[Union[int, Literal["latest"], Literal["pending"], Literal["earliest"], Literal["safe"], Literal["finalized"]]] = None, confirmations: Optional[int] = None) -> {returns_str}:""",
            1,
        )
        line, _ = self.__get_line_pos_from_byte_offset(
            declaration.file, declaration.byte_location[0]
        )
        self.add_str_to_types(2, '"""', 1)
        self.add_str_to_types(
            2,
            f"[Source code]({_path_to_uri(declaration.file)}#{line + 1})",
            1 if len(param_names) + len(returns) == 0 else 2,
        )
        if len(param_names) > 0:
            self.add_str_to_types(2, "Args:", 1)
            for param_name, param_type in param_names:
                self.add_str_to_types(3, f"{param_name}: {param_type}", 1)
        if len(returns) == 1:
            self.add_str_to_types(2, "Returns:", 1)
            self.add_str_to_types(3, f"{returns[0][1]}", 1)
        elif len(returns) > 1:
            self.add_str_to_types(2, "Returns:", 1)
            self.add_str_to_types(3, f'({", ".join(ret[1] for ret in returns)})', 1)
        self.add_str_to_types(2, '"""', 1)
        self.add_str_to_types(2, "...", 2)

    def generate_types_function(self, fn: FunctionDefinition):
        params_names, params = self.generate_func_params(fn)
        returns = self.generate_func_returns(fn)

        if len(returns) == 0:
            returns_str = "None"
        elif len(returns) == 1:
            returns_str = returns[0][0]
        else:
            returns_str = f"Tuple[{', '.join(ret[0] for ret in returns)}]"

        is_pure_or_view = fn.state_mutability in {
            StateMutability.PURE,
            StateMutability.VIEW,
        }

        self.generate_type_hint_stub_func(
            fn,
            params,
            returns_str,
            "call",
            is_pure_or_view,
            params_names,
            returns,
        )
        self.generate_type_hint_stub_func(
            fn,
            params,
            "int",
            "estimate",
            False,
            params_names,
            returns,
        )
        self.generate_type_hint_stub_func(
            fn,
            params,
            "Tuple[Dict[Address, List[int]], int]",
            "access_list",
            False,
            params_names,
            returns,
        )
        self.generate_type_hint_stub_func(
            fn,
            params,
            f"TransactionAbc[{returns_str}]",
            "tx",
            not is_pure_or_view,
            params_names,
            returns,
        )

        assert fn.function_selector is not None
        self.generate_func_implementation(
            fn,
            params,
            params_names,
            returns,
        )

    def generate_types_contract(self, contract: ContractDefinition) -> None:
        fqn = f"{contract.parent.source_unit_name}:{contract.name}"
        if fqn in self.__already_generated_contracts:
            return
        else:
            self.__already_generated_contracts.add(fqn)

        base_names: List[str] = []

        for base in reversed(contract.base_contracts):
            parent_contract = base.base_name.referenced_declaration
            assert isinstance(parent_contract, ContractDefinition)
            base_names.append(self.get_name(parent_contract))
            # only the types for contracts in the same source_unit are generated
            if (
                parent_contract.parent.source_unit_name
                == contract.parent.source_unit_name
            ):
                self.generate_types_contract(parent_contract)
            # contract is not in the same source unit, so it must be imported
            else:
                self.__imports.generate_contract_import(parent_contract, force=True)

        contract_module_name = "pytypes." + _make_path_alphanum(
            contract.parent.source_unit_name[:-3]
        ).replace("/", ".")
        self.__contracts_index[fqn] = (
            contract_module_name,
            (self.get_name(contract),),
        )

        if len(base_names) == 0:
            base_names = ["Contract"]

        events_abi = self.generate_contract_template(contract, ", ".join(base_names))

        if contract.enums:
            self.generate_types_enum(contract.enums, 1)

        if contract.structs:
            self.generate_types_struct(contract.structs, 1)

        if contract.errors:
            self.generate_types_error(contract.errors, 1)

        if contract.events:
            self.generate_types_event(contract.events, 1, events_abi)

        selector_assignments = []

        for var in contract.declared_variables:
            if (
                var.visibility == Visibility.EXTERNAL
                or var.visibility == Visibility.PUBLIC
            ):
                self.generate_getter_for_state_var(var)
                selector_assignments.append((self.get_name(var), var.function_selector))
        for fn in contract.functions:
            if fn.function_selector:
                if contract.kind != ContractKind.LIBRARY or fn.state_mutability in {
                    StateMutability.VIEW,
                    StateMutability.PURE,
                }:
                    self.generate_types_function(fn)
                    selector_assignments.append(
                        (self.get_name(fn), fn.function_selector)
                    )

        for fn_name, selector in selector_assignments:
            self.add_str_to_types(
                0, f"{self.get_name(contract)}.{fn_name}.selector = {selector}", 1
            )

    def generate_types_source_unit(self, unit: SourceUnit) -> None:
        self.generate_types_struct(unit.structs, 0)
        self.generate_types_enum(unit.enums, 0)
        self.generate_types_error(unit.errors, 0)
        for contract in unit.contracts:
            self.generate_types_contract(contract)

    def clean_type_dir(self):
        """
        instead of recursive removal of type files inside pytypes dir
        remove the root and recreate it
        """
        if self.__pytypes_dir.exists():
            shutil.rmtree(self.__pytypes_dir)
        self.__pytypes_dir.mkdir(exist_ok=True)

    def write_source_unit_to_file(self, contract_name: str):
        self.__pytypes_dir.mkdir(exist_ok=True)
        contract_name = _make_path_alphanum(contract_name[:-3])
        unit_path = (self.__pytypes_dir / contract_name).with_suffix(".py")
        unit_path_parent = unit_path.parent
        # TODO validate whether project root can become paraent
        unit_path_parent.mkdir(parents=True, exist_ok=True)
        if unit_path.exists():
            with unit_path.open("a") as f:
                f.write(str(self.__imports) + self.__source_unit_types)
        else:
            unit_path.touch()
            unit_path.write_text(str(self.__imports) + self.__source_unit_types)

    # clean the instance variables to enable generating a new source unit
    def cleanup_source_unit(self):
        self.__source_unit_types = ""
        self.__imports.cleanup_imports()

    def add_func_overload_if_match(
        self, fn: FunctionDefinition, contract: ContractDefinition
    ):
        for function in contract.functions:
            # this function is called also with the contract in which the function is defined
            # thus we need to ensure that we don't compare the function with itself
            if (
                fn.name == function.name and fn != function
            ):  # and len(fn.parameters.parameters) == len(function.parameters.parameters):
                # both functions have to be overloded -> add both
                if isinstance(fn.parent, ContractDefinition):
                    source_unit = fn.parent.parent
                else:
                    source_unit = fn.parent
                # there can be 2 contracts with the same name and both of them can define function with the same name
                # thus to uniquely idenify the funtion also the source unit has to be used, otherwise it could happen
                # that an incorrect function gets overloaded
                self.__func_to_overload.add(
                    source_unit.source_unit_name + fn.canonical_name
                )
                self.__func_to_overload.add(
                    contract.parent.source_unit_name + function.canonical_name
                )
                # print("-------------------")
                # print(f"overload: {fn.canonical_name} {fn.parent.canonical_name}")
                # print(f"overload: {function.canonical_name} {contract.canonical_name}")
                # print("-------------------")

    # TODO add check if func not in __func_to_overload for optimization
    def traverse_funcs_in_child_contracts(
        self, fn: FunctionDefinition, contract: ContractDefinition
    ):
        for child in contract.child_contracts:
            self.add_func_overload_if_match(fn, child)

        for child in contract.child_contracts:
            self.traverse_funcs_in_child_contracts(fn, child)

    # TODO add check if func not in __func_to_overload
    def traverse_funcs_in_parent_contracts(
        self, fn: FunctionDefinition, contract: ContractDefinition
    ):
        for inh_spec in contract.base_contracts:
            if not contract.kind == ContractKind.INTERFACE:
                parent_contract = inh_spec.base_name.referenced_declaration
                assert isinstance(parent_contract, ContractDefinition)
                self.add_func_overload_if_match(fn, parent_contract)
        for inh_spec in contract.base_contracts:
            if not contract.kind == ContractKind.INTERFACE:
                parent_contract = inh_spec.base_name.referenced_declaration
                assert isinstance(parent_contract, ContractDefinition)
                self.traverse_funcs_in_parent_contracts(fn, parent_contract)

    # TODO travesrse also state variables as getters are generated for them and thus overlaoding might be necessary
    def traverse_funcs_to_check_overload(self):
        # set containing canonical names of functions to be overloaded
        for _, unit in self.__source_units.items():
            for contract in unit.contracts:
                # interface function can't have implementation and thus can't be overloaded
                if contract.kind != ContractKind.INTERFACE:
                    for fn in contract.functions:
                        if (
                            not fn.canonical_name in self.__func_to_overload
                            and fn.implemented
                            and fn.function_selector
                        ):
                            # we create pytypes only for functions that are publicly accessible
                            if (
                                fn.visibility == Visibility.PUBLIC
                                or fn.visibility == Visibility.EXTERNAL
                            ):
                                self.add_func_overload_if_match(fn, contract)
                                self.traverse_funcs_in_parent_contracts(fn, contract)
                                self.traverse_funcs_in_child_contracts(fn, contract)
            # if unit.source_unit_name == "overloading.sol":
            #    print(self.__func_to_overload)

    def generate_types(self, compiler: SolidityCompiler) -> None:
        def generate_source_unit(source_unit: SourceUnit) -> None:
            self.__current_source_unit = source_unit.source_unit_name
            self.generate_types_source_unit(source_unit)
            self.write_source_unit_to_file(source_unit.source_unit_name)
            self.cleanup_source_unit()
            self.__name_sanitizer.clear_global_renames()

        build = compiler.latest_build
        assert build is not None
        self.__interval_trees = build.interval_trees
        self.__source_units = build.source_units
        self.__reference_resolver = build.reference_resolver

        self.clean_type_dir()

        # generate source units in import order, source units with no imports are generated first
        # also handle cyclic imports
        assert compiler.latest_graph is not None
        assert compiler.latest_source_units_to_paths
        graph: nx.DiGraph = (
            compiler.latest_graph.copy()
        )  # pyright: ignore reportGeneralTypeIssues
        source_units_to_paths = compiler.latest_source_units_to_paths
        paths_to_source_unit_names: DefaultDict[Path, Set[str]] = defaultdict(set)
        for source_unit, path in source_units_to_paths.items():
            paths_to_source_unit_names[path].add(source_unit)

        previous_len = len(graph)
        cycles_detected = False
        cycles: Set[FrozenSet[str]] = set()

        for cycle in nx.simple_cycles(graph):
            for source_unit_name in cycle:
                self.__cyclic_source_units[source_unit_name].update(
                    [s for s in cycle if s != source_unit_name]
                )

        while len(graph) > 0:
            # use heapq to make order of source units deterministic
            sources: List[str] = [
                node
                for node, in_degree in graph.in_degree()  # pyright: ignore reportGeneralTypeIssues
                if in_degree == 0
            ]
            heapq.heapify(sources)
            visited: Set[str] = set(sources)

            while len(sources) > 0:
                source = heapq.heappop(sources)
                path = source_units_to_paths[source]
                if path in self.__source_units:
                    generate_source_unit(self.__source_units[path])

                for source_unit_name in paths_to_source_unit_names[path]:
                    visited.add(source_unit_name)
                    for (
                        _,
                        to,
                    ) in graph.out_edges(  # pyright: ignore reportGeneralTypeIssues
                        source_unit_name  # pyright: ignore reportGeneralTypeIssues
                    ):
                        if (
                            graph.in_degree(
                                to  # pyright: ignore reportGeneralTypeIssues
                            )
                            == 1
                        ):
                            heapq.heappush(sources, to)
                            visited.add(to)
                graph.remove_nodes_from(paths_to_source_unit_names[path])

            generated_cycles: Set[FrozenSet[str]] = set()

            for cycle in nx.simple_cycles(graph):
                cycles_detected = True
                cycles.add(frozenset(cycle))
                if frozenset(cycle) in generated_cycles:
                    continue

                is_closed_cycle = True
                for node in cycle:
                    if any(
                        edge[0] not in cycle
                        for edge in graph.in_edges(
                            node  # pyright: ignore reportGeneralTypeIssues
                        )
                    ):
                        is_closed_cycle = False
                        break

                if is_closed_cycle:
                    generated_cycles.add(frozenset(cycle))

            for cycle in sorted(generated_cycles):
                for source in cycle:
                    path = source_units_to_paths[source]
                    if path in self.__source_units:
                        generate_source_unit(self.__source_units[path])
                    graph.remove_nodes_from(paths_to_source_unit_names[path])

            if len(graph) == previous_len:
                break
            previous_len = len(graph)

        if cycles_detected:
            logger.warning(
                "Cyclic imports detected, pytypes may not be working correctly:\n"
                + "\n".join(str(set(cycle)) for cycle in cycles)
            )

        init_path = self.__pytypes_dir / "__init__.py"
        init_path.write_text(
            INIT_CONTENT.format(
                version=get_package_version("woke"),
                errors=self.__errors_index,
                events=self.__events_index,
                contracts_by_fqn=self.__contracts_index,
                contracts_by_metadata=self.__contracts_by_metadata_index,
                contracts_inheritance=self.__contracts_inheritance_index,
                contracts_revert_index=self.__contracts_revert_index,
                creation_code_index=self.__creation_code_index,
            )
        )


class SourceUnitImports:
    __all_imports: str
    __struct_imports: Set[str]
    __enum_imports: Set[str]
    __contract_imports: Set[str]
    __python_imports: Set[str]
    __type_checking_imports: Set[str]
    __tmp_type_checking_imports: Set[str]
    __type_gen: TypeGenerator

    def __init__(self, outer: TypeGenerator):
        self.__struct_imports = set()
        self.__enum_imports = set()
        self.__all_imports = ""
        self.__contract_imports = set()
        self.__python_imports = set()
        self.__type_checking_imports = set()
        self.__tmp_type_checking_imports = set()
        self.__type_gen = outer

    def __str__(self) -> str:
        self.__add_str_to_imports(0, DEFAULT_IMPORTS, 1)

        for python_import in sorted(self.__python_imports):
            self.__add_str_to_imports(0, python_import, 1)

        if self.__python_imports:
            self.__add_str_to_imports(0, "", 1)

        for contract in sorted(self.__contract_imports):
            self.__add_str_to_imports(0, contract, 1)

        if self.__contract_imports:
            self.__add_str_to_imports(0, "", 1)

        for struct in sorted(self.__struct_imports):
            self.__add_str_to_imports(0, struct, 1)

        if self.__struct_imports:
            self.__add_str_to_imports(0, "", 1)

        for enum in sorted(self.__enum_imports):
            self.__add_str_to_imports(0, enum, 1)

        if self.__enum_imports:
            self.__add_str_to_imports(0, "", 1)

        if self.__type_checking_imports:
            self.__add_str_to_imports(0, "from typing import TYPE_CHECKING", 1)
            self.__add_str_to_imports(0, "if TYPE_CHECKING:", 1)
            for type_checking_import in sorted(self.__type_checking_imports):
                self.__add_str_to_imports(1, type_checking_import, 1)
            self.__add_str_to_imports(0, "", 1)

        self.__add_str_to_imports(0, "", 2)

        return self.__all_imports

    def cleanup_imports(self) -> None:
        self.__struct_imports.clear()
        self.__enum_imports.clear()
        self.__contract_imports.clear()
        self.__python_imports.clear()
        self.__type_checking_imports.clear()
        self.__all_imports = ""

    def clear_tmp_type_checking_imports(self) -> None:
        self.__tmp_type_checking_imports.clear()

    @property
    def tmp_type_checking_imports(self) -> FrozenSet[str]:
        return frozenset(self.__tmp_type_checking_imports)

    def __generate_import(
        self, declaration: DeclarationAbc, source_unit_name: str
    ) -> str:
        source_unit_name = _make_path_alphanum(source_unit_name)
        return (
            "from pytypes."
            + source_unit_name[:-3].replace("/", ".")
            + " import "
            + self.__type_gen.get_name(declaration)
        )

    def __add_str_to_imports(
        self, num_of_indentation: int, string: str, num_of_newlines: int
    ):
        self.__all_imports += (
            num_of_indentation * TAB_WIDTH * " " + string + num_of_newlines * "\n"
        )

    def generate_struct_import(self, struct_type: types.Struct):
        node = struct_type.ir_node
        if isinstance(node.parent, ContractDefinition):
            source_unit = node.parent.parent
        else:
            source_unit = node.parent
        # only those structs that are defined in a different source unit should be imported
        if source_unit.source_unit_name == self.__type_gen.current_source_unit:
            return
        struct_import = self.__generate_import(
            struct_type.ir_node, source_unit.source_unit_name
        )

        if source_unit.source_unit_name in self.__type_gen.cyclic_source_units:
            self.__type_checking_imports.add(struct_import)
            self.__tmp_type_checking_imports.add(struct_import)
        else:
            self.__struct_imports.add(struct_import)

    # only used for top-level enums (not within contracts)
    def generate_enum_import(self, enum_type: types.Enum):
        source_unit = enum_type.ir_node.parent
        assert isinstance(source_unit, SourceUnit)

        # only those structs that are defined in a different source unit should be imported
        if source_unit.source_unit_name == self.__type_gen.current_source_unit:
            return
        enum_import = self.__generate_import(
            enum_type.ir_node, source_unit.source_unit_name
        )

        if source_unit.source_unit_name in self.__type_gen.cyclic_source_units:
            self.__type_checking_imports.add(enum_import)
            self.__tmp_type_checking_imports.add(enum_import)
        else:
            self.__enum_imports.add(enum_import)

    def generate_contract_import(
        self, contract: ContractDefinition, *, force: bool = False
    ):
        source_unit = contract.parent
        if source_unit.source_unit_name == self.__type_gen.current_source_unit:
            return

        contract_import = self.__generate_import(contract, source_unit.source_unit_name)

        if (
            not force
            and contract_import not in self.__contract_imports
            and source_unit.source_unit_name in self.__type_gen.cyclic_source_units
        ):
            self.__type_checking_imports.add(contract_import)
            self.__tmp_type_checking_imports.add(contract_import)
        else:
            self.__contract_imports.add(contract_import)

    def add_python_import(self, p_import: str) -> None:
        self.__python_imports.add(p_import)


class NameSanitizer:
    __global_reserved: Set[str]
    __contract_reserved: Set[str]
    __function_reserved: Set[str]
    __struct_reserved: Set[str]
    __event_reserved: Set[str]
    __error_reserved: Set[str]
    __enum_reserved: Set[str]

    __global_renames: Dict[DeclarationAbc, str]
    __contract_renames: DefaultDict[ContractDefinition, Dict[DeclarationAbc, str]]
    __function_renames: DefaultDict[FunctionDefinition, Dict[DeclarationAbc, str]]
    __struct_renames: DefaultDict[StructDefinition, Dict[DeclarationAbc, str]]
    __event_renames: DefaultDict[EventDefinition, Dict[DeclarationAbc, str]]
    __error_renames: DefaultDict[ErrorDefinition, Dict[DeclarationAbc, str]]
    __enum_renames: DefaultDict[EnumDefinition, Dict[DeclarationAbc, str]]

    def __init__(self):
        self.__global_reserved = {
            "Dict",
            "List",
            "Mapping",
            "Set",
            "Tuple",
            "Union",
            "Annotated",
            "Optional",
            "Literal",
            "Callable",
            "Path",
            "bytearray",
            "IntEnum",
            "dataclasses",
            "overload",
            "Contract",
            "Library",
            "Address",
            "Account",
            "Chain",
            "RequestType",
            "TransactionRevertedError",
            "TransactionAbc",
            "LegacyTransaction",
            "Eip2930Transaction",
            "Eip1559Transaction",
            "ValueRange",
            "Length",
            "bytes",
            "int",
            "uint",
            "str",
            "bool",
        }

        for i in range(8, 257, 8):
            self.__global_reserved.add(f"uint{i}")
            self.__global_reserved.add(f"int{i}")

        for i in range(1, 33):
            self.__global_reserved.add(f"bytes{i}")
            self.__global_reserved.add(f"List{i}")

        self.__contract_reserved = {
            "_abi",
            "_creation_code",
            "_address",
            "_chain",
            "_label",
            "_get_creation_code",
            "_deploy",
            "_execute",
            "_library_id",
            "_prepare_tx_params",
            "address",
            "label",
            "balance",
            "code",
            "chain",
            "nonce",
            "call",
            "transact",
            "estimate",
            "deploy",
            "get_creation_code",
        }
        self.__function_reserved = {
            "self",
            "cls",
            "from_",
            "value",
            "gas_limit",
            "return_tx",
            "chain",
            "to",
            "request_type",
            "gas_price",
            "max_fee_per_gas",
            "max_priority_fee_per_gas",
            "access_list",
            "block",
            "confirmations",
        }
        self.__struct_reserved = {"original_name"}
        self.__event_reserved = {"_abi", "selector", "original_name"}
        self.__error_reserved = {"_abi", "selector", "original_name"}
        self.__enum_reserved = set()

        self.__global_renames = {}
        self.__contract_renames = defaultdict(dict)
        self.__function_renames = defaultdict(dict)
        self.__struct_renames = defaultdict(dict)
        self.__event_renames = defaultdict(dict)
        self.__error_renames = defaultdict(dict)
        self.__enum_renames = defaultdict(dict)

    def _check_global(self, name: str) -> bool:
        return (
            name in self.__global_reserved
            or name in set(self.__global_renames.values())
            or keyword.iskeyword(name)
        )

    def _check_contract(self, name: str, contract: ContractDefinition) -> bool:
        return (
            name in self.__global_reserved
            or name in set(self.__global_renames)
            or name in self.__contract_reserved
            or name in set(self.__contract_renames[contract].values())
            or keyword.iskeyword(name)
            or (
                name.startswith("__")
                and name.endswith("__")
                and not name.endswith("___")
            )
        )

    def _check_function(self, name: str, function: FunctionDefinition) -> bool:
        return (
            name in self.__global_reserved
            or name in set(self.__global_renames)
            or name in self.__function_reserved
            or name in set(self.__function_renames[function].values())
            or keyword.iskeyword(name)
        )

    def _check_struct(self, name: str, struct: StructDefinition) -> bool:
        return (
            name in self.__global_reserved
            or name in set(self.__global_renames)
            or name in self.__struct_reserved
            or name in set(self.__struct_renames[struct].values())
            or keyword.iskeyword(name)
            or (
                name.startswith("__")
                and name.endswith("__")
                and not name.endswith("___")
            )
        )

    def _check_event(self, name: str, event: EventDefinition) -> bool:
        return (
            name in self.__global_reserved
            or name in set(self.__global_renames)
            or name in self.__event_reserved
            or name in set(self.__event_renames[event].values())
            or keyword.iskeyword(name)
            or (
                name.startswith("__")
                and name.endswith("__")
                and not name.endswith("___")
            )
        )

    def _check_error(self, name: str, error: ErrorDefinition) -> bool:
        return (
            name in self.__global_reserved
            or name in set(self.__global_renames)
            or name in self.__error_reserved
            or name in set(self.__error_renames[error].values())
            or keyword.iskeyword(name)
            or (
                name.startswith("__")
                and name.endswith("__")
                and not name.endswith("___")
            )
        )

    def _check_enum(self, name: str, enum: EnumDefinition) -> bool:
        return (
            name in self.__global_reserved
            or name in set(self.__global_renames)
            or name in self.__enum_reserved
            or name in set(self.__enum_renames[enum].values())
            or keyword.iskeyword(name)
            or (
                name.startswith("__")
                and name.endswith("__")
                and not name.endswith("___")
            )
        )

    def clear_global_renames(self):
        self.__global_renames = {}

    def sanitize_name(self, declaration: DeclarationAbc) -> str:
        parent = declaration.parent
        if isinstance(parent, SourceUnit):
            check = self._check_global
            renames = self.__global_renames
        elif isinstance(parent, ContractDefinition):
            check = lambda name: self._check_contract(name, parent)
            renames = self.__contract_renames[parent]
        elif isinstance(parent, StructDefinition):
            check = lambda name: self._check_struct(name, parent)
            renames = self.__struct_renames[parent]
        elif isinstance(parent, EnumDefinition):
            check = lambda name: self._check_enum(name, parent)
            renames = self.__enum_renames[parent]
        elif isinstance(parent, ParameterList):
            parent_parent = parent.parent
            if isinstance(parent_parent, FunctionDefinition):
                check = lambda name: self._check_function(name, parent_parent)
                renames = self.__function_renames[parent_parent]
            elif isinstance(parent_parent, EventDefinition):
                check = lambda name: self._check_event(name, parent_parent)
                renames = self.__event_renames[parent_parent]
            elif isinstance(parent_parent, ErrorDefinition):
                check = lambda name: self._check_error(name, parent_parent)
                renames = self.__error_renames[parent_parent]
            else:
                raise NotImplementedError(
                    f"Cannot sanitize name for declaration {declaration} with parent {parent} and parent of parent {parent_parent}"
                )
        else:
            raise NotImplementedError(
                f"Cannot sanitize name for declaration {declaration} with parent {parent}"
            )

        if declaration in renames:
            return renames[declaration]

        new_name = declaration.name
        while new_name.startswith("__"):
            new_name = new_name[1:]

        while check(new_name):
            new_name = new_name + "_"

        renames[declaration] = new_name
        return new_name



from typing import Any


class LspError(Exception):
    __code: int
    __message: str
    __data: Any

    def __init__(self, code: int, message: str, data: Any = None):
        self.__code = code
        self.__message = message
        self.__data = data

    @property
    def code(self) -> int:
        return self.__code

    @property
    def message(self) -> str:
        return self.__message

    @property
    def data(self) -> Any:
        return self.__data



from __future__ import annotations

import re
from collections import deque
from functools import lru_cache, partial
from typing import (
    TYPE_CHECKING,
    Deque,
    FrozenSet,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from ..meta.override_specifier import OverrideSpecifier
from ..reference_resolver import CallbackParams
from ..statement.block import Block
from .abc import DeclarationAbc

if TYPE_CHECKING:
    from woke.analysis.cfg import ControlFlowGraph
    from .contract_definition import ContractDefinition
    from ..expression.identifier import Identifier
    from ..expression.member_access import MemberAccess
    from ..meta.identifier_path import IdentifierPathPart
    from ..statement.inline_assembly import ExternalReference
    from ..expression.unary_operation import UnaryOperation
    from ..expression.binary_operation import BinaryOperation

from woke.ast.enums import Visibility
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.meta.parameter_list import ParameterList
from woke.ast.ir.meta.structured_documentation import StructuredDocumentation
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import (
    AstNodeId,
    SolcModifierDefinition,
    SolcStructuredDocumentation,
)


class ModifierDefinition(DeclarationAbc):
    """
    Definition of a modifier.

    !!! example
        ```solidity
        modifier onlyOwner {
            require(msg.sender == owner);
            _;
        }
        ```
    """

    _ast_node: SolcModifierDefinition
    _parent: ContractDefinition
    _child_modifiers: Set[ModifierDefinition]

    _body: Optional[Block]
    _implemented: bool
    _parameters: ParameterList
    _virtual: bool
    _visibility: Visibility
    _base_modifiers: List[AstNodeId]
    _documentation: Optional[Union[StructuredDocumentation, str]]
    _overrides: Optional[OverrideSpecifier]

    def __init__(
        self, init: IrInitTuple, modifier: SolcModifierDefinition, parent: SolidityAbc
    ):
        super().__init__(init, modifier, parent)
        self._child_modifiers = set()

        self._body = Block(init, modifier.body, self) if modifier.body else None
        self._implemented = self._body is not None
        self._parameters = ParameterList(init, modifier.parameters, self)
        self._virtual = modifier.virtual
        self._visibility = modifier.visibility
        self._base_modifiers = (
            list(modifier.base_modifiers) if modifier.base_modifiers is not None else []
        )
        if modifier.documentation is None:
            self._documentation = None
        elif isinstance(modifier.documentation, SolcStructuredDocumentation):
            self._documentation = StructuredDocumentation(
                init, modifier.documentation, self
            )
        elif isinstance(modifier.documentation, str):
            self._documentation = modifier.documentation
        else:
            raise TypeError(
                f"Unknown type of documentation: {type(modifier.documentation)}"
            )
        self._overrides = (
            OverrideSpecifier(init, modifier.overrides, self)
            if modifier.overrides is not None
            else None
        )
        self._reference_resolver.register_post_process_callback(self._post_process)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        if self._body is not None:
            yield from self._body
        yield from self._parameters
        if isinstance(self._documentation, StructuredDocumentation):
            yield from self._documentation
        if self._overrides is not None:
            yield from self._overrides

    def _post_process(self, callback_params: CallbackParams):
        base_modifiers = self.base_modifiers
        for base_modifier in base_modifiers:
            base_modifier._child_modifiers.add(self)
        self._reference_resolver.register_destroy_callback(
            self.file, partial(self._destroy, base_modifiers)
        )

    def _destroy(self, base_modifiers: Tuple[ModifierDefinition]) -> None:
        for base_modifier in base_modifiers:
            base_modifier._child_modifiers.remove(self)

    def _parse_name_location(self) -> Tuple[int, int]:
        IDENTIFIER = r"[a-zA-Z$_][a-zA-Z0-9$_]*"
        MODIFIER_RE = re.compile(
            r"^\s*modifier\s+(?P<name>{identifier})".format(
                identifier=IDENTIFIER
            ).encode("utf-8")
        )

        byte_start = self._ast_node.src.byte_offset
        match = MODIFIER_RE.match(self._source)
        assert match
        return byte_start + match.start("name"), byte_start + match.end("name")

    def get_all_references(
        self, include_declarations: bool
    ) -> Iterator[
        Union[
            DeclarationAbc,
            Identifier,
            IdentifierPathPart,
            MemberAccess,
            ExternalReference,
            UnaryOperation,
            BinaryOperation,
        ]
    ]:
        processed_declarations: Set[ModifierDefinition] = {self}
        declarations_queue: Deque[ModifierDefinition] = deque([self])

        while declarations_queue:
            declaration = declarations_queue.pop()
            if include_declarations:
                yield declaration
            yield from declaration.references

            for base_modifier in declaration.base_modifiers:
                if base_modifier not in processed_declarations:
                    declarations_queue.append(base_modifier)
                    processed_declarations.add(base_modifier)
            for child_modifier in declaration.child_modifiers:
                if child_modifier not in processed_declarations:
                    declarations_queue.append(child_modifier)
                    processed_declarations.add(child_modifier)

    @property
    def parent(self) -> ContractDefinition:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def canonical_name(self) -> str:
        return f"{self._parent.canonical_name}.{self._name}"

    @property
    @lru_cache(maxsize=2048)
    def declaration_string(self) -> str:
        ret = f"modifier {self._name}"
        ret += (
            f"({', '.join(param.declaration_string for param in self.parameters.parameters)})"
            if len(self.parameters.parameters) > 0
            else ""
        )
        ret += " virtual" if self.virtual else ""
        ret += (
            (
                f" override"
                + (
                    "("
                    + ", ".join(
                        override.source for override in self.overrides.overrides
                    )
                    + ")"
                    if len(self.overrides.overrides) > 0
                    else ""
                )
            )
            if self.overrides is not None
            else ""
        )

        if isinstance(self.documentation, StructuredDocumentation):
            return (
                "/// "
                + "\n///".join(line for line in self.documentation.text.splitlines())
                + "\n"
                + ret
            )
        elif isinstance(self.documentation, str):
            return (
                "/// "
                + "\n///".join(line for line in self.documentation.splitlines())
                + "\n"
                + ret
            )
        else:
            return ret

    @property
    def body(self) -> Optional[Block]:
        """
        Returns:
            Body of the modifier, if any.
        """
        return self._body

    @property
    def implemented(self) -> bool:
        """
        Returns:
            `True` if the modifier [body][woke.ast.ir.declaration.modifier_definition.ModifierDefinition.body] is not `None`, `False` otherwise.
        """
        return self._implemented

    @property
    def parameters(self) -> ParameterList:
        """
        Returns:
            Parameter list describing the modifier parameters.
        """
        return self._parameters

    @property
    def virtual(self) -> bool:
        """
        Returns:
            `True` if the modifier is virtual, `False` otherwise.
        """
        return self._virtual

    @property
    def visibility(self) -> Visibility:
        """
        Returns:
            Visibility of the modifier.
        """
        return self._visibility

    @property
    def base_modifiers(self) -> Tuple[ModifierDefinition, ...]:
        """
        !!! example
            `B.mod` on lines 6-8 lists `A.mod` on line 2 as a base modifier.

            `C.mod` on lines 12-14 lists only `B.mod` on lines 6-8 as a base modifier.
            ```solidity linenums="1"
            abstract contract A {
                modifier mod virtual;
            }

            contract B is A {
                modifier mod virtual override {
                    _;
                }
            }

            contract C is B {
                modifier mod override {
                    _;
                }
            }
            ```

        !!! example
            `B1.mod` on lines 6-8 lists `A.mod` on line 2 as a base modifier.

            `B2.mod` on lines 12-14 lists `A.mod` on line 2 as a base modifier.

            `C.mod` on lines 18-20 lists `B1.mod` on lines 6-8 and `B2.mod` on lines 12-14 as base modifiers.
            ```solidity linenums="1"
            abstract contract A {
                modifier mod virtual;
            }

            contract B1 is A {
                modifier mod virtual override {
                    _;
                }
            }

            contract B2 is A {
                modifier mod virtual override {
                    _;
                }
            }

            contract C is B1, B2 {
                modifier mod override(B1, B2) {
                    _;
                }
            }
            ```
        Returns:
            List of base modifiers overridden by this modifier.
        """
        base_modifiers = []
        for base_modifier_id in self._base_modifiers:
            base_modifier = self._reference_resolver.resolve_node(
                base_modifier_id, self._cu_hash
            )
            assert isinstance(base_modifier, ModifierDefinition)
            base_modifiers.append(base_modifier)
        return tuple(base_modifiers)

    @property
    def child_modifiers(self) -> FrozenSet[ModifierDefinition]:
        """
        Returns:
            Modifiers that list this modifier in their [base_modifiers][woke.ast.ir.declaration.modifier_definition.ModifierDefinition.base_modifiers] property.
        """
        return frozenset(self._child_modifiers)

    @property
    def documentation(self) -> Optional[Union[StructuredDocumentation, str]]:
        """
        Of [StructuredDocumentation][woke.ast.ir.meta.structured_documentation.StructuredDocumentation] type since Solidity 0.6.3.
        Returns:
            [NatSpec](https://docs.soliditylang.org/en/latest/natspec-format.html) documentation string, if any.
        """
        return self._documentation

    @property
    def overrides(self) -> Optional[OverrideSpecifier]:
        """
        Returns override specifier as present in the source code.
        !!! example
            `A.mod` on line 2 does not have an override specifier.

            `B1.mod` on lines 6-8 has an override specifier with the [overrides][woke.ast.ir.meta.override_specifier.OverrideSpecifier.overrides] property empty.

            `B2.mod` on lines 12-14 has an override specifier with the [overrides][woke.ast.ir.meta.override_specifier.OverrideSpecifier.overrides] property empty.

            `C.mod` on lines 18-20 has an override specifier with the [overrides][woke.ast.ir.meta.override_specifier.OverrideSpecifier.overrides] property containg two items referencing the contracts `B1` and `B2` ([ContractDefinition][woke.ast.ir.declaration.contract_definition.ContractDefinition]).
            ```solidity linenums="1"
            abstract contract A {
                modifier mod virtual;
            }

            contract B1 is A {
                modifier mod virtual override {
                    _;
                }
            }

            contract B2 is A {
                modifier mod virtual override {
                    _;
                }
            }

            contract C is B1, B2 {
                modifier mod override(B1, B2) {
                    _;
                }
            }
            ```

        Returns:
            Override specifier, if any.
        """
        return self._overrides

    @property
    @lru_cache(maxsize=2048)
    def cfg(self) -> Optional[ControlFlowGraph]:
        from woke.analysis.cfg import ControlFlowGraph

        if self.body is None:
            return None
        return ControlFlowGraph(self)



import logging
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Optional, Set, Tuple

from woke.ast.enums import ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import (
    SolcAssignment,
    SolcBinaryOperation,
    SolcConditional,
    SolcElementaryTypeNameExpression,
    SolcExpressionUnion,
    SolcFunctionCall,
    SolcFunctionCallOptions,
    SolcIdentifier,
    SolcIndexAccess,
    SolcIndexRangeAccess,
    SolcLiteral,
    SolcMemberAccess,
    SolcNewExpression,
    SolcTupleExpression,
    SolcUnaryOperation,
    TypeDescriptionsModel,
)
from woke.ast.types import TypeAbc
from woke.utils.string import StringReader

logger = logging.getLogger(__name__)


class ExpressionAbc(SolidityAbc, ABC):
    """
    Abstract base class for all IR expression nodes.
    > Something that has a value.
    """

    _type_descriptions: TypeDescriptionsModel

    def __init__(
        self, init: IrInitTuple, expression: SolcExpressionUnion, parent: SolidityAbc
    ):
        super().__init__(init, expression, parent)
        self._type_descriptions = expression.type_descriptions

    @staticmethod
    def from_ast(
        init: IrInitTuple, expression: SolcExpressionUnion, parent: SolidityAbc
    ) -> "ExpressionAbc":
        from .assignment import Assignment
        from .binary_operation import BinaryOperation
        from .conditional import Conditional
        from .elementary_type_name_expression import ElementaryTypeNameExpression
        from .function_call import FunctionCall
        from .function_call_options import FunctionCallOptions
        from .identifier import Identifier
        from .index_access import IndexAccess
        from .index_range_access import IndexRangeAccess
        from .literal import Literal
        from .member_access import MemberAccess
        from .new_expression import NewExpression
        from .tuple_expression import TupleExpression
        from .unary_operation import UnaryOperation

        if isinstance(expression, SolcAssignment):
            return Assignment(init, expression, parent)
        elif isinstance(expression, SolcBinaryOperation):
            return BinaryOperation(init, expression, parent)
        elif isinstance(expression, SolcConditional):
            return Conditional(init, expression, parent)
        elif isinstance(expression, SolcElementaryTypeNameExpression):
            return ElementaryTypeNameExpression(init, expression, parent)
        elif isinstance(expression, SolcFunctionCall):
            return FunctionCall(init, expression, parent)
        elif isinstance(expression, SolcFunctionCallOptions):
            return FunctionCallOptions(init, expression, parent)
        elif isinstance(expression, SolcIdentifier):
            return Identifier(init, expression, parent)
        elif isinstance(expression, SolcIndexAccess):
            return IndexAccess(init, expression, parent)
        elif isinstance(expression, SolcIndexRangeAccess):
            return IndexRangeAccess(init, expression, parent)
        elif isinstance(expression, SolcLiteral):
            return Literal(init, expression, parent)
        elif isinstance(expression, SolcMemberAccess):
            return MemberAccess(init, expression, parent)
        elif isinstance(expression, SolcNewExpression):
            return NewExpression(init, expression, parent)
        elif isinstance(expression, SolcTupleExpression):
            return TupleExpression(init, expression, parent)
        elif isinstance(expression, SolcUnaryOperation):
            return UnaryOperation(init, expression, parent)

    @property
    @lru_cache(maxsize=2048)
    def type(self) -> Optional[TypeAbc]:
        """
        Can be `None` in case of an [Identifier][woke.ast.ir.expression.identifier.Identifier] in an [ImportDirective][woke.ast.ir.meta.import_directive.ImportDirective].
        !!! example
            `Ownable` in the following example has no type information:
            ```solidity
            import { Ownable } from './Ownable.sol';
            ```
        Returns:
            Type of the expression.
        """
        if self._type_descriptions.type_identifier is None:
            return None

        type_identifier = StringReader(self._type_descriptions.type_identifier)
        ret = TypeAbc.from_type_identifier(
            type_identifier, self._reference_resolver, self.cu_hash
        )
        assert (
            len(type_identifier) == 0
        ), f"Failed to parse type_identifier: {self._type_descriptions.type_identifier}"
        return ret

    @property
    def type_string(self) -> Optional[str]:
        """
        !!! example
            `:::solidity function (uint256,uint256) returns (uint256)` in the case of the `foo` [Identifier][woke.ast.ir.expression.identifier.Identifier] in the `:::solidity foo(1, 2)` expression for the following function:
            ```solidity
            function foo(uint a, uint b) public onlyOwner payable virtual onlyOwner returns(uint) {
                return a + b;
            }
            ```

        Can be `None` in case of an [Identifier][woke.ast.ir.expression.identifier.Identifier] in an [ImportDirective][woke.ast.ir.meta.import_directive.ImportDirective].
        !!! example
            `Ownable` in the following example has no type information:
            ```solidity
            import { Ownable } from './Ownable.sol';
            ```
        Returns:
            User-friendly string describing the expression type.
        """
        return self._type_descriptions.type_string

    @property
    @abstractmethod
    def is_ref_to_state_variable(self) -> bool:
        """
        In many cases it may be useful to know if an [Assignment][woke.ast.ir.expression.assignment.Assignment] to an expression modifies a state variable or not.
        This may not be straightforward to determine, e.g. if the expression is a [MemberAccess][woke.ast.ir.expression.member_access.MemberAccess] or [IndexAccess][woke.ast.ir.expression.index_access.IndexAccess] to a state variable.
        Returns:
            `True` if the expression (possibly) is a reference to a state variable.
        """
        ...

    @property
    @abstractmethod
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        """
        Returns:
            Set of child IR nodes (including `self`) that modify the blockchain state and flags describing how the state is modified.
        """
        ...



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Tuple, Union

from ..type_name.user_defined_type_name import UserDefinedTypeName
from .identifier_path import IdentifierPath

if TYPE_CHECKING:
    from ..declaration.function_definition import FunctionDefinition
    from ..declaration.modifier_definition import ModifierDefinition
    from ..declaration.variable_declaration import VariableDeclaration

from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import (
    SolcIdentifierPath,
    SolcOverrideSpecifier,
    SolcUserDefinedTypeName,
)


class OverrideSpecifier(SolidityAbc):
    """
    !!! example
        An override specifier can be used:

        - in a [FunctionDefinition][woke.ast.ir.declaration.function_definition.FunctionDefinition]:
            - `:::solidity override` in line 19,
        - in a [ModifierDefinition][woke.ast.ir.declaration.modifier_definition.ModifierDefinition]:
            - `:::solidity override` in line 12,
        - in a [VariableDeclaration][woke.ast.ir.declaration.variable_declaration.VariableDeclaration]:
            - `:::solidity override(IERC20)` in line 17.

        ```solidity linenums="1"
        interface IERC20 {
            function transfer(address to, uint256 value) external returns (bool);

            function allowance(address owner, address spender) external view returns (uint256);
        }

        abstract contract ERC20 is IERC20 {
            modifier EOA() virtual;
        }

        contract C is ERC20 {
            modifier EOA() override {
                require(msg.sender == tx.origin);
                _;
            }

            mapping(address => mapping(address => uint256)) public override(IERC20) allowance;

            function transfer(address to, uint256 value) external override returns (bool) {
                // ...
            }
        }
        ```
    """

    _ast_node: SolcOverrideSpecifier
    _parent: Union[FunctionDefinition, ModifierDefinition, VariableDeclaration]

    _overrides: List[Union[IdentifierPath, UserDefinedTypeName]]

    def __init__(
        self,
        init: IrInitTuple,
        override_specifier: SolcOverrideSpecifier,
        parent: SolidityAbc,
    ):
        super().__init__(init, override_specifier, parent)
        self._overrides = []

        for override in override_specifier.overrides:
            if isinstance(override, SolcIdentifierPath):
                self._overrides.append(IdentifierPath(init, override, self))
            elif isinstance(override, SolcUserDefinedTypeName):
                self._overrides.append(UserDefinedTypeName(init, override, self))

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for override in self._overrides:
            yield from override

    @property
    def parent(
        self,
    ) -> Union[FunctionDefinition, ModifierDefinition, VariableDeclaration]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def overrides(self) -> Tuple[Union[IdentifierPath, UserDefinedTypeName], ...]:
        """
        !!! note
            Is empty when there are no curly braces after the `:::solidity override` keyword.
        Returns:
            Tuple of IR nodes referencing the contract or interface whose declaration is being overridden.
        """
        return tuple(self._overrides)



import sys
from abc import abstractmethod

# inspired by https://github.com/algrebe/python-tee/blob/master/tee/tee.py


class Tee:
    def __init__(self, filename, mode="a"):
        self.filename = filename
        self.mode = mode

        self.stream = None
        self.fp = None

    @abstractmethod
    def set_stream(self, stream):
        pass

    @abstractmethod
    def get_stream(self):
        pass

    def write(self, message):
        self.stream.write(message)  # pyright: ignore reportGeneralTypeIssues
        self.fp.write(message)  # pyright: ignore reportOptionalMemberAccess

    def flush(self):
        self.stream.flush()  # pyright: ignore reportGeneralTypeIssues
        self.fp.flush()  # pyright: ignore reportOptionalMemberAccess

    def __enter__(self):
        self.stream = self.get_stream()
        self.fp = open(self.filename, self.mode)
        self.set_stream(self)

    def __exit__(self, *args):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        if self.stream is not None:
            self.set_stream(self.stream)
            self.stream = None

        if self.fp is not None:
            self.fp.close()
            self.fp = None

    def isatty(self):
        return self.stream.isatty()  # pyright: ignore reportGeneralTypeIssues

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.filename)

    __str__ = __repr__
    __unicode__ = __repr__


class StdoutTee(Tee):
    def set_stream(self, stream):
        sys.stdout = stream

    def get_stream(self):
        return sys.stdout


class StderrTee(Tee):
    def set_stream(self, stream):
        sys.stderr = stream

    def get_stream(self):
        return sys.stderr



class ChecksumError(Exception):
    """
    Checksum of a downloaded file did not match the expected value.
    """


class UnsupportedVersionError(Exception):
    """
    The provided version is lower than the minimal supported version on the current platform.
    """



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, Optional, Set, Tuple, Union

from woke.ast.enums import ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.expression.abc import ExpressionAbc
from woke.ast.ir.statement.abc import StatementAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcIfStatement

if TYPE_CHECKING:
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .unchecked_block import UncheckedBlock
    from .while_statement import WhileStatement


class IfStatement(StatementAbc):
    """
    !!! example
        Lines 2-6 in the following code:
        ```solidity linenums="1"
        function foo(int x) public pure returns(uint) {
            if (x < 0) {
                return 0;
            } else {
                return uint(x);
            }
        }
        ```
    """

    _ast_node: SolcIfStatement
    _parent: Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]

    _condition: ExpressionAbc
    _true_body: StatementAbc
    _false_body: Optional[StatementAbc]

    def __init__(
        self, init: IrInitTuple, if_statement: SolcIfStatement, parent: SolidityAbc
    ):
        super().__init__(init, if_statement, parent)
        self._condition = ExpressionAbc.from_ast(init, if_statement.condition, self)
        self._true_body = StatementAbc.from_ast(init, if_statement.true_body, self)
        self._false_body = (
            None
            if if_statement.false_body is None
            else StatementAbc.from_ast(init, if_statement.false_body, self)
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._condition
        yield from self._true_body
        if self._false_body is not None:
            yield from self._false_body

    @property
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def condition(self) -> ExpressionAbc:
        """
        Returns:
            Condition of the if statement.
        """
        return self._condition

    @property
    def true_body(self) -> StatementAbc:
        """
        Returns:
            Statement executed if the condition is true.
        """
        return self._true_body

    @property
    def false_body(self) -> Optional[StatementAbc]:
        """
        Returns:
            Statement executed if the condition is false (if any).
        """
        return self._false_body

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return (
            self.condition.modifies_state
            | self.true_body.modifies_state
            | (self.false_body.modifies_state if self.false_body is not None else set())
        )

    def statements_iter(self) -> Iterator["StatementAbc"]:
        yield self
        yield from self._true_body.statements_iter()
        if self._false_body is not None:
            yield from self._false_body.statements_iter()



from typing import Set

from intervaltree import Interval

from woke.lsp.common_structures import Position, Range


def changes_to_byte_offset(changes: Set[Interval]) -> int:
    byte_offset = 0
    tag: str
    j1: int
    j2: int
    for change in changes:
        tag, j1, j2 = change.data
        if tag == "insert":
            byte_offset += j2 - j1 - 1
        elif tag == "delete":
            byte_offset -= change.end - change.begin - 1
        elif tag == "replace":
            byte_offset += j2 - j1 - (change.end - change.begin)
        else:
            raise ValueError(f"Unknown tag {tag}")
    return byte_offset


def position_within_range(pos: Position, range: Range) -> bool:
    if pos.line < range.start.line:
        return False
    if pos.line == range.start.line and pos.character < range.start.character:
        return False
    if pos.line > range.end.line:
        return False
    if pos.line == range.end.line and pos.character > range.end.character:
        return False
    return True



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union

from woke.ast.ir.meta.identifier_path import IdentifierPath
from woke.ast.ir.type_name.abc import TypeNameAbc
from woke.ast.ir.type_name.user_defined_type_name import UserDefinedTypeName
from woke.ast.ir.utils import IrInitTuple

from ...enums import BinaryOpOperator, UnaryOpOperator

if TYPE_CHECKING:
    from woke.ast.ir.declaration.contract_definition import ContractDefinition
    from .source_unit import SourceUnit

from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.nodes import (
    SolcIdentifierPath,
    SolcUserDefinedTypeName,
    SolcUsingForDirective,
)


class UsingForDirective(SolidityAbc):
    """
    !!! note
        Either [library_name][woke.ast.ir.meta.using_for_directive.UsingForDirective.library_name] must be set or one of [functions][woke.ast.ir.meta.using_for_directive.UsingForDirective.functions] or [operator_functions][woke.ast.ir.meta.using_for_directive.UsingForDirective.operator_functions] must be non-empty.
    !!! example
        Lines 13 and 14 in the following example:
        ```solidity linenums="1"
        type I8 is int8;

        function add(uint a, uint b) pure returns (uint) {
            return a + b;
        }

        function sub(I8 a, I8 b) pure returns (I8) {
            return I8.wrap(I8.unwrap(a) - I8.unwrap(b));
        }

        library SafeMath {
            function sub(uint a, uint b) public pure returns (uint) {
                require(b <= a);
                return a - b;
            }
        }

        using {sub as -} for I8 global;

        contract C {
            using SafeMath for uint;
            using {add} for uint;
        }
        ```
    """

    _ast_node: SolcUsingForDirective
    _parent: Union[ContractDefinition, SourceUnit]

    _functions: List[IdentifierPath]
    _operator_functions: List[
        Tuple[IdentifierPath, Union[UnaryOpOperator, BinaryOpOperator]]
    ]
    _library_name: Optional[Union[IdentifierPath, UserDefinedTypeName]]
    _type_name: Optional[TypeNameAbc]
    # TODO add _global

    def __init__(
        self,
        init: IrInitTuple,
        using_for_directive: SolcUsingForDirective,
        parent: Union[ContractDefinition, SourceUnit],
    ):
        super().__init__(init, using_for_directive, parent)

        if using_for_directive.function_list is None:
            self._functions = []
            self._operator_functions = []
        else:
            self._functions = [
                IdentifierPath(init, function.function, self)
                for function in using_for_directive.function_list
                if function.function is not None
            ]
            self._operator_functions = [
                (
                    IdentifierPath(init, function.definition, self),
                    function.operator,
                )
                for function in using_for_directive.function_list
                if function.definition is not None and function.operator is not None
            ]

        if using_for_directive.library_name is None:
            self._library_name = None
        elif isinstance(using_for_directive.library_name, SolcUserDefinedTypeName):
            self._library_name = UserDefinedTypeName(
                init, using_for_directive.library_name, self
            )
        elif isinstance(using_for_directive.library_name, SolcIdentifierPath):
            self._library_name = IdentifierPath(
                init, using_for_directive.library_name, self
            )

        if using_for_directive.type_name is None:
            self._type_name = None
        else:
            self._type_name = TypeNameAbc.from_ast(
                init, using_for_directive.type_name, self
            )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for function in self._functions:
            yield from function
        for function, _ in self._operator_functions:
            yield from function
        if self._library_name is not None:
            yield from self._library_name
        if self._type_name is not None:
            yield from self._type_name

    @property
    def parent(self) -> Union[ContractDefinition, SourceUnit]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def functions(self) -> Tuple[IdentifierPath, ...]:
        """
        Returns:
            List of functions that are attached to the target type.
        """
        return tuple(self._functions)

    @property
    def operator_functions(
        self,
    ) -> Tuple[Tuple[IdentifierPath, Union[UnaryOpOperator, BinaryOpOperator]], ...]:
        """
        Returns:
            List of operator functions and their operators that are attached to the target type.
        """
        return tuple(self._operator_functions)

    @property
    def library_name(self) -> Optional[Union[IdentifierPath, UserDefinedTypeName]]:
        """
        Is only set in the case of `:::solidity using LibraryName for TypeName;` directive type.
        Returns:
            IR node referencing the library ([ContractDefinition][woke.ast.ir.declaration.contract_definition.ContractDefinition] of the [ContractKind.LIBRARY][woke.ast.enums.ContractKind.LIBRARY] kind) that is attached to the target type.
        """
        return self._library_name

    @property
    def type_name(self) -> Optional[TypeNameAbc]:
        """
        Is `None` in the case of `:::solidity using Lib for *;`.
        Returns:
            Type name that is attached to the functions or library.
        """
        return self._type_name



from typing import List, Set

from woke.analysis.detectors import DetectorAbc, DetectorResult, detector
from woke.ast.enums import ContractKind
from woke.ast.ir.declaration.contract_definition import ContractDefinition


@detector(-1040, "unused-contract")
class UnusedContractDetector(DetectorAbc):
    """
    Detects abstract contracts, interfaces and libraries that are not used.
    """

    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def report(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_contract_definition(self, node: ContractDefinition):
        if len(node.references) == 0:
            if node.abstract:
                self._detections.add(
                    DetectorResult(
                        node, "Contract not used", lsp_range=node.name_location
                    )
                )
            elif node.kind == ContractKind.INTERFACE:
                self._detections.add(
                    DetectorResult(
                        node, "Interface not used", lsp_range=node.name_location
                    )
                )
            elif node.kind == ContractKind.LIBRARY:
                self._detections.add(
                    DetectorResult(
                        node, "Library not used", lsp_range=node.name_location
                    )
                )



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, List, Optional, Set, Tuple, Union

from woke.ast.enums import ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.ast.ir.expression.abc import ExpressionAbc
from woke.ast.ir.statement.abc import StatementAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import AstNodeId, SolcVariableDeclarationStatement

if TYPE_CHECKING:
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .if_statement import IfStatement
    from .unchecked_block import UncheckedBlock
    from .while_statement import WhileStatement


class VariableDeclarationStatement(StatementAbc):
    """
    !!! example
        `:::solidity (uint a, uint b) = (1, 2)` in the following code:
        ```solidity
        contract C {
            function f() public {
                (uint a, uint b) = (1, 2);
            }
        }
        ```
    """

    _ast_node: SolcVariableDeclarationStatement
    _parent: Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]

    _assignments: List[Optional[AstNodeId]]
    _declarations: List[Optional[VariableDeclaration]]
    _initial_value: Optional[ExpressionAbc]

    def __init__(
        self,
        init: IrInitTuple,
        variable_declaration_statement: SolcVariableDeclarationStatement,
        parent: SolidityAbc,
    ):
        super().__init__(init, variable_declaration_statement, parent)
        self._assignments = list(variable_declaration_statement.assignments)

        self._declarations = []
        for declaration in variable_declaration_statement.declarations:
            if declaration is None:
                self._declarations.append(None)
            else:
                self._declarations.append(VariableDeclaration(init, declaration, self))

        if variable_declaration_statement.initial_value is None:
            self._initial_value = None
        else:
            self._initial_value = ExpressionAbc.from_ast(
                init, variable_declaration_statement.initial_value, self
            )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for declaration in self._declarations:
            if declaration is not None:
                yield from declaration
        if self._initial_value is not None:
            yield from self._initial_value

    @property
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def declarations(self) -> Tuple[Optional[VariableDeclaration], ...]:
        """
        !!! example
            Some declarations may be `None`, e.g. in the following code:
            ```solidity
            (bool success, ) = address(this).call{value: 1}("");
            ```

        Returns:
            Tuple of variable declarations in this statement.
        """
        return tuple(self._declarations)

    @property
    def initial_value(self) -> Optional[ExpressionAbc]:
        """
        Does not need to be a [TupleExpression][woke.ast.ir.expression.tuple_expression.TupleExpression] when there is more than one variable declared.
        Can also be a [FunctionCall][woke.ast.ir.expression.function_call.FunctionCall] returning a tuple.
        Returns:
            Initial value assigned to the declared variables (if any).
        """
        return self._initial_value

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        ret = set()
        if self.initial_value is not None:
            ret |= self.initial_value.modifies_state
            if any(
                declaration.is_state_variable
                for declaration in self.declarations
                if declaration is not None
            ):
                ret |= {(self, ModifiesStateFlag.MODIFIES_STATE_VAR)}
        return ret



from __future__ import annotations

import inspect
import json
import math
from collections import namedtuple
from functools import lru_cache
from typing import Callable, Dict, Iterable, List, Optional, Tuple, TypeVar, Union
from urllib.error import URLError
from urllib.parse import urlparse
from urllib.request import Request, urlopen

from Crypto.Hash import keccak
from eth_utils.abi import function_abi_to_4byte_selector

from ..utils import get_package_version
from .core import Account, Address
from .globals import get_config

ChainExplorer = namedtuple("ChainExplorer", ["url", "api_url"])

chain_explorer_urls: Dict[int, ChainExplorer] = {
    1: ChainExplorer("https://etherscan.io", "https://api.etherscan.io/api"),
    5: ChainExplorer(
        "https://goerli.etherscan.io", "https://api-goerli.etherscan.io/api"
    ),
    56: ChainExplorer("https://bscscan.com", "https://api.bscscan.com/api"),
    97: ChainExplorer(
        "https://testnet.bscscan.com", "https://api-testnet.bscscan.com/api"
    ),
    137: ChainExplorer("https://polygonscan.com", "https://api.polygonscan.com/api"),
    80001: ChainExplorer(
        "https://mumbai.polygonscan.com", "https://api-mumbai.polygonscan.com/api"
    ),
    43114: ChainExplorer("https://snowtrace.io/", "https://api.snowtrace.io/api"),
    43113: ChainExplorer(
        "https://testnet.snowtrace.io/", "https://api-testnet.snowtrace.io/api"
    ),
    10: ChainExplorer(
        "https://optimistic.etherscan.io/", "https://api-optimistic.etherscan.io/api"
    ),
    420: ChainExplorer(
        "https://goerli-optimism.etherscan.io/",
        "https://api-goerli-optimism.etherscan.io/api",
    ),
    100: ChainExplorer("https://gnosisscan.io/", "https://api.gnosisscan.io/api"),
    42161: ChainExplorer("https://arbiscan.io/", "https://api.arbiscan.io/api"),
    421613: ChainExplorer(
        "https://testnet.arbiscan.io/", "https://api-testnet.arbiscan.io/api"
    ),
    84531: ChainExplorer(
        "https://goerli.basescan.org/", "https://api-goerli.basescan.org/api"
    ),
    11155111: ChainExplorer(
        "https://sepolia.etherscan.io/", "https://api-sepolia.etherscan.io/api"
    ),
    1101: ChainExplorer(
        "https://zkevm.polygonscan.com/", "https://api-zkevm.polygonscan.com/api"
    ),
    1442: ChainExplorer(
        "https://testnet-zkevm.polygonscan.com/",
        "https://api-testnet-zkevm.polygonscan.com/api",
    ),
    42220: ChainExplorer(
        "https://celoscan.io/",
        "https://api.celoscan.io/api",
    ),
    44787: ChainExplorer(
        "https://alfajores.celoscan.io/",
        "https://api-alfajores.celoscan.io/api",
    ),
    1284: ChainExplorer(
        "https://moonscan.io/",
        "https://api-moonbeam.moonscan.io/api",
    ),
    1287: ChainExplorer(
        "https://moonbase.moonscan.io/",
        "https://api-moonbase.moonscan.io/api",
    ),
    250: ChainExplorer(
        "https://ftmscan.com/",
        "https://api.ftmscan.com/api",
    ),
    4002: ChainExplorer(
        "https://testnet.ftmscan.com/",
        "https://api-testnet.ftmscan.com/api",
    ),
}


@lru_cache(maxsize=1024)
def get_contract_info_from_explorer(
    addr: Address, chain_id: int
) -> Optional[Tuple[str, Dict]]:
    if chain_id not in chain_explorer_urls:
        return None

    u = urlparse(chain_explorer_urls[chain_id].url)
    config = get_config()
    api_key = config.api_keys.get(".".join(u.netloc.split(".")[:-1]), None)
    if api_key is None:
        return None

    url = (
        chain_explorer_urls[chain_id].api_url
        + f"?module=contract&action=getsourcecode&address={addr}&apikey={api_key}"
    )

    req = Request(
        url,
        headers={
            "Content-Type": "application/json",
            "User-Agent": f"woke/{get_package_version('woke')}",
        },
    )

    try:
        with urlopen(req) as response:
            ret = json.loads(response.read().decode("utf-8"))
    except URLError as e:
        return None

    if ret["status"] != "1":
        return None

    data = ret["result"][0]
    if data["ContractName"] == "":
        return None

    abi = {}
    # TODO library ABI is different and has to be fixed to compute the correct selector
    # however, it is not possible to detect if a contract is a library or not without parsing the source code
    for abi_item in json.loads(data["ABI"]):
        if abi_item["type"] in {"constructor", "fallback", "receive"}:
            abi[abi_item["type"]] = abi_item
        elif abi_item["type"] == "function":
            abi[function_abi_to_4byte_selector(abi_item)] = abi_item

    return data["ContractName"], abi


def format_int(x: int) -> str:
    if abs(x) < 10**5:
        return f"{x:_}"
    no_of_digits = int(math.log10(abs(x))) + 1
    if x % 10 ** (no_of_digits - 3) == 0:
        return f"{x:.2e}"
    return f"{x:.2E} ({x:_})"


def exists(a) -> bool:
    if a is None:
        return False
    if len(a) >= 0:
        return True
    return bool(a)


T = TypeVar("T")


def partition(
    seq: Iterable[T], predicate: Callable[[T], bool]
) -> Tuple[List[T], List[T]]:
    p = []
    not_p = []
    for el in seq:
        if predicate(el):
            p.append(el)
        else:
            not_p.append(el)
    return p, not_p


def keccak256(b: bytes) -> bytes:
    h = keccak.new(data=b, digest_bits=256)
    return h.digest()


def get_current_fn_name(back_count=1) -> str:
    frame = inspect.currentframe()
    assert frame is not None
    for _ in range(back_count):
        frame = frame.f_back
        assert frame is not None
    return frame.f_code.co_name


def negate(fn):
    def inner(*args, **kwargs) -> bool:
        return not fn(*args, **kwargs)

    return inner


def get_create_address(deployer: Union[Account, Address, str], nonce: int) -> Address:
    if isinstance(deployer, Account):
        deployer = deployer.address
    deployer_bytes = bytes.fromhex(str(deployer)[2:])

    # see https://ethereum.org/en/developers/docs/data-structures-and-encoding/rlp
    if nonce < 0:
        raise ValueError("Nonce must be positive")
    elif nonce == 0:
        data = b"\xd6\x94" + deployer_bytes + b"\x80"
    elif nonce <= 0x7F:
        data = b"\xd6\x94" + deployer_bytes + bytes([nonce])
    elif nonce <= 0xFF:
        data = b"\xd7\x94" + deployer_bytes + b"\x81" + bytes([nonce])
    elif nonce <= 0xFFFF:
        data = b"\xd8\x94" + deployer_bytes + b"\x82" + nonce.to_bytes(2, "big")
    elif nonce <= 0xFFFFFF:
        data = b"\xd9\x94" + deployer_bytes + b"\x83" + nonce.to_bytes(3, "big")
    elif nonce <= 0xFFFFFFFF:
        data = b"\xda\x94" + deployer_bytes + b"\x84" + nonce.to_bytes(4, "big")
    else:
        raise ValueError("Nonce too large")

    return Address("0x" + keccak256(data)[-20:].hex())


def get_create2_address_from_hash(
    deployer: Union[Account, Address, str], salt: bytes, creation_code_hash: bytes
) -> Address:
    if isinstance(deployer, Account):
        deployer = deployer.address
    deployer_bytes = bytes.fromhex(str(deployer)[2:])

    return Address(
        "0x"
        + keccak256(b"\xff" + deployer_bytes + salt + creation_code_hash)[-20:].hex()
    )


def get_create2_address_from_code(
    deployer: Union[Account, Address, str], salt: bytes, creation_code: bytes
) -> Address:
    return get_create2_address_from_hash(deployer, salt, keccak256(creation_code))



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, Tuple

if TYPE_CHECKING:
    from woke.ast.ir.meta.source_unit import SourceUnit

from woke.ast.nodes import SolcPragmaDirective

from ..abc import SolidityAbc
from ..utils import IrInitTuple

logger = logging.getLogger(__name__)


class PragmaDirective(SolidityAbc):
    _ast_node: SolcPragmaDirective
    _parent: SourceUnit

    _literals: List[str]

    def __init__(
        self, init: IrInitTuple, pragma: SolcPragmaDirective, parent: SolidityAbc
    ):
        super().__init__(init, pragma, parent)
        self._literals = list(pragma.literals)

    @property
    def parent(self) -> SourceUnit:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def literals(self) -> Tuple[str, ...]:
        """
        !!! example
            `:::py ('solidity', '^', '0.8', '||', '0.7', '.1', '-', '0.7', '.6')` for the following pragma:
            ```solidity
            pragma solidity ^0.8 || 0.7.1 - 0.7.6;
            ```
        !!! example
            `:::py ('abicoder', 'v2')` for the following pragma:
            ```solidity
            pragma abicoder v2;
            ```
        !!! example
            `:::py ('experimental', 'SMTChecker')` for the following pragma:
            ```solidity
            pragma experimental SMTChecker;
            ```

        Returns:
            Literals of the pragma directive.
        """
        return tuple(self._literals)



import logging
from typing import List, Sequence, Set

import woke.ast.types as types
from woke.analysis.detectors import DetectorAbc, DetectorResult, detector
from woke.ast.enums import DataLocation, FunctionCallKind, FunctionTypeKind, Visibility
from woke.ast.ir.abc import IrAbc
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.expression.function_call_options import FunctionCallOptions
from woke.ast.ir.expression.identifier import Identifier
from woke.ast.ir.expression.member_access import MemberAccess
from woke.ast.ir.expression.new_expression import NewExpression
from woke.ast.ir.expression.tuple_expression import TupleExpression
from woke.ast.ir.meta.source_unit import SourceUnit
from woke.core.solidity_version import SolidityVersionRange, SolidityVersionRanges

logger = logging.getLogger(__name__)


def type_contains_dynamic_component(t: types.TypeAbc) -> bool:
    if isinstance(t, types.Array):
        if t.length is None:
            return True
        return type_contains_dynamic_component(t.base_type)
    elif isinstance(t, (types.Bytes, types.String)):
        return True
    elif isinstance(t, types.Tuple):
        return any(
            type_contains_dynamic_component(component)
            for component in t.components
            if component is not None
        )
    elif isinstance(t, types.Struct):
        return any(type_contains_dynamic_component(m.type) for m in t.ir_node.members)
    else:
        return False


def abi_encoder_v2_enabled(node: IrAbc) -> bool:
    source_unit = node
    while source_unit is not None:
        if isinstance(source_unit, SourceUnit):
            break
        source_unit = source_unit.parent
    assert isinstance(source_unit, SourceUnit)

    if any(
        directive.literals == ("abicoder", "v1") for directive in source_unit.pragmas
    ):
        return False
    elif any(
        directive.literals == ("abicoder", "v2") for directive in source_unit.pragmas
    ):
        return True
    elif any(
        directive.literals == ("experimental", "ABIEncoderV2")
        for directive in source_unit.pragmas
    ):
        return True
    else:
        default_v2_versions = SolidityVersionRanges(
            [SolidityVersionRange("0.8.0", True, None, None)]
        )
        if len(default_v2_versions & node.version_ranges) == 0:
            return False
        return True


def types_meet_requirements(
    t: Sequence[types.TypeAbc], data_location_is_calldata: bool = False
) -> bool:
    for component in t:
        if isinstance(component, types.Struct):
            bug = types_meet_requirements(
                list(m.type for m in component.ir_node.members),
                data_location_is_calldata
                or component.data_location == DataLocation.CALLDATA,
            )
            if bug:
                return True

    if len(t) < 2:
        return False
    last = t[-1]
    if not (
        isinstance(last, types.Array)
        and last.length is not None
        and (last.data_location == DataLocation.CALLDATA or data_location_is_calldata)
    ):
        return False

    base_type = last.base_type
    while isinstance(base_type, types.Array):
        if base_type.length is None:
            return False
        base_type = base_type.base_type
    if isinstance(base_type, types.UInt) and base_type.bits_count == 256:
        pass
    elif isinstance(base_type, types.FixedBytes) and base_type.bytes_count == 32:
        pass
    else:
        return False

    if not any(type_contains_dynamic_component(x) for x in t[:-1]):
        return False

    return True


@detector(-1020, "overflow-calldata-tuple-reencoding-bug")
class OverflowCalldataTupleReencodingBugDetector(DetectorAbc):
    """
    Detects Head Overflow Calldata Tuple Reencoding compiler bug
    """

    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def report(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_function_definition(self, node: FunctionDefinition):
        if node.visibility not in {Visibility.PUBLIC, Visibility.EXTERNAL}:
            return None

        versions = node.version_ranges
        affected_versions = SolidityVersionRanges(
            [SolidityVersionRange("0.5.8", True, "0.8.16", False)]
        )
        if len(versions & affected_versions) == 0:
            return

        # ABI encoder v2 must be enabled
        if not abi_encoder_v2_enabled(node):
            return

        encoded_types = [param.type for param in node.return_parameters.parameters]
        if types_meet_requirements(encoded_types):
            self._detections.add(
                DetectorResult(
                    node.return_parameters,
                    "Found head overflow calldata tuple reencoding compiler bug",
                )
            )

    def visit_function_call(self, node: FunctionCall):
        if node.kind == FunctionCallKind.TYPE_CONVERSION:
            return

        versions = node.version_ranges
        affected_versions = SolidityVersionRanges(
            [SolidityVersionRange("0.5.8", True, "0.8.16", False)]
        )
        if len(versions & affected_versions) == 0:
            return

        func_identifier = node.expression
        while True:
            if isinstance(func_identifier, (Identifier, MemberAccess)):
                break
            elif isinstance(func_identifier, FunctionCallOptions):
                func_identifier = func_identifier.expression
            elif isinstance(func_identifier, NewExpression):
                return
            elif (
                isinstance(func_identifier, TupleExpression)
                and len(func_identifier.components) == 1
            ):
                func_identifier = func_identifier.components[0]
            elif isinstance(func_identifier, FunctionCall):
                t = func_identifier.type
                if isinstance(t, types.Function) and (t.value_set or t.gas_set):
                    func_identifier = func_identifier.expression
                else:
                    # function returning a function
                    break
            else:
                logger.warning(
                    f"Unexpected function call child node: {func_identifier} {func_identifier.source if func_identifier else ''}"
                )
                return

        t = func_identifier.type

        if not isinstance(t, types.Function):
            return

        if t.kind == FunctionTypeKind.EXTERNAL:
            assert t.attached_to is None
            encoded_types = [arg.type for arg in node.arguments]
        elif t.kind in {FunctionTypeKind.ERROR, FunctionTypeKind.EVENT}:
            encoded_types = [arg.type for arg in node.arguments]
        elif t.kind in {
            FunctionTypeKind.ABI_ENCODE,
            FunctionTypeKind.ABI_ENCODE_PACKED,
        }:
            encoded_types = [arg.type for arg in node.arguments]
        elif t.kind in {
            FunctionTypeKind.ABI_ENCODE_WITH_SELECTOR,
            FunctionTypeKind.ABI_ENCODE_WITH_SIGNATURE,
        }:
            assert len(node.arguments) >= 1
            encoded_types = [arg.type for arg in node.arguments[1:]]
        elif t.kind == FunctionTypeKind.ABI_ENCODE_CALL:
            assert len(node.arguments) == 2
            if isinstance(node.arguments[1], TupleExpression):
                encoded_types = [
                    arg.type if arg is not None else None
                    for arg in node.arguments[
                        1
                    ].components  # pyright: ignore reportGeneralTypeIssues
                ]
            elif (
                isinstance(node.arguments[1], FunctionCall)
                and node.arguments[1].kind == FunctionCallKind.STRUCT_CONSTRUCTOR_CALL
            ):
                # probably always created in memory, not calldata
                return
            else:
                logger.warning(
                    f"Unexpected abi.encodeCall second argument: {node.arguments[1]}"
                )
                return
        else:
            return

        # ABI encoder v2 must be enabled
        if not abi_encoder_v2_enabled(node):
            return

        if types_meet_requirements(
            encoded_types  # pyright: ignore reportGeneralTypeIssues
        ):
            self._detections.add(
                DetectorResult(
                    node, "Found head overflow calldata tuple reencoding compiler bug"
                )
            )



from __future__ import annotations

from functools import lru_cache, reduce
from operator import or_
from typing import TYPE_CHECKING, Iterator, List, Set, Tuple, Union

from woke.ast.enums import ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.meta.try_catch_clause import TryCatchClause
from woke.ast.ir.statement.abc import StatementAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcTryStatement

if TYPE_CHECKING:
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .if_statement import IfStatement
    from .unchecked_block import UncheckedBlock
    from .while_statement import WhileStatement


class TryStatement(StatementAbc):
    """
    !!! example
        ```solidity
        try this.f() returns (uint256) {
            // ...
        } catch Error(string memory reason) {
            // ...
        } catch Panic(uint errorCode) {
            // ...
        } catch (bytes memory lowLevelData) {
            // ...
        }
        ```
    """

    _ast_node: SolcTryStatement
    _parent: Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]

    _clauses: List[TryCatchClause]
    _external_call: FunctionCall

    def __init__(
        self, init: IrInitTuple, try_statement: SolcTryStatement, parent: SolidityAbc
    ):
        super().__init__(init, try_statement, parent)
        self._clauses = [
            TryCatchClause(init, clause, self) for clause in try_statement.clauses
        ]
        self._external_call = FunctionCall(init, try_statement.external_call, self)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for clause in self._clauses:
            yield from clause
        yield from self._external_call

    @property
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def clauses(self) -> Tuple[TryCatchClause, ...]:
        """
        Returns:
            Try/catch clauses.
        """
        return tuple(self._clauses)

    @property
    def external_call(self) -> FunctionCall:
        """
        Returns:
            External call executed in the try statement.
        """
        return self._external_call

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return (
            reduce(
                or_,
                (clause.modifies_state for clause in self._clauses),
                set(),
            )
            | self.external_call.modifies_state
        )

    def statements_iter(self) -> Iterator["StatementAbc"]:
        yield self
        for clause in self._clauses:
            yield from clause.block.statements_iter()



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Tuple, Union

from woke.core.solidity_version import VersionAbc


class CompilerVersionManagerAbc(ABC):
    """
    ABC for all compiler version managers.
    """

    @abstractmethod
    def installed(self, version: Union[VersionAbc, str]) -> bool:
        """
        Check if a compiler version is installed.
        :param version: compiler version to check
        :return: True if installed, False otherwise
        """

    @abstractmethod
    async def install(
        self, version: Union[VersionAbc, str], force_reinstall: bool = False
    ) -> None:
        """
        Install the target compiler version.
        :param version: compiler version to be installed
        :param force_reinstall: if True, download target compiler version even if already installed
        """

    @abstractmethod
    def remove(self, version: Union[VersionAbc, str]) -> None:
        """
        Remove the target compiler version.
        :param version: compiler version to be removed
        """

    @abstractmethod
    def get_path(self, version: Union[VersionAbc, str]) -> Path:
        """
        Return a system path of the target compiler version executable.
        :param version: version of the compiler executable whose path is returned
        """

    @abstractmethod
    def list_all(self) -> Tuple[VersionAbc]:
        """
        Return a set of all supported compiler versions.
        :return set of all supported compiler versions
        """

    def list_installed(self) -> Tuple[VersionAbc, ...]:
        """
        Return a set of installed compiler versions.
        :return: set of installed compiler versions
        """
        return tuple(version for version in self.list_all() if self.installed(version))



import logging
import os
import platform
import reprlib
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, FrozenSet, Iterable, Optional, Set, Tuple, Union

import networkx as nx
import tomli

from woke.utils import change_cwd

from ..core.solidity_version import SolidityVersion
from .data_model import (
    CompilerConfig,
    DeploymentConfig,
    DetectorsConfig,
    GeneratorConfig,
    LspConfig,
    TestingConfig,
    TopLevelConfig,
)

logger = logging.getLogger(__name__)


class UnsupportedPlatformError(Exception):
    """
    The current platform is not supported. Supported platforms are: Linux, macOS, Windows.
    """


class WokeConfig:
    __project_root_path: Path
    __global_config_path: Path
    __global_data_path: Path
    __loaded_files: Set[Path]
    __config_raw: Dict[str, Any]
    __config: TopLevelConfig

    def __init__(
        self,
        *_,
        project_root_path: Optional[Union[str, Path]] = None,
    ):
        system = platform.system()

        try:
            self.__global_config_path = Path(os.environ["XDG_CONFIG_HOME"]) / "woke"
        except KeyError:
            if system in {"Linux", "Darwin"}:
                self.__global_config_path = Path.home() / ".config" / "woke"
            elif system == "Windows":
                self.__global_config_path = Path(os.environ["LOCALAPPDATA"]) / "woke"
            else:
                raise UnsupportedPlatformError(f"Platform `{system}` is not supported.")

        try:
            self.__global_data_path = Path(os.environ["XDG_DATA_HOME"]) / "woke"
        except KeyError:
            if system in {"Linux", "Darwin"}:
                self.__global_data_path = Path.home() / ".local" / "share" / "woke"
            elif system == "Windows":
                self.__global_data_path = Path(os.environ["LOCALAPPDATA"]) / "woke"
            else:
                raise UnsupportedPlatformError(f"Platform `{system}` is not supported.")

        migrate = False
        if (
            not self.__global_config_path.exists()
            and not self.__global_data_path.exists()
        ):
            migrate = True

        self.__global_config_path.mkdir(parents=True, exist_ok=True)
        self.__global_data_path.mkdir(parents=True, exist_ok=True)

        if migrate:
            if system == "Linux":
                old_path = Path.home() / ".config" / "Woke"
            elif system == "Darwin":
                old_path = Path.home() / ".config" / "Woke"
            elif system == "Windows":
                old_path = Path.home() / "Woke"
            else:
                raise UnsupportedPlatformError(f"Platform `{system}` is not supported.")

            config_path = old_path / "config.toml"
            compilers_path = old_path / "compilers"
            solc_versions_path = old_path / ".woke_solc_version"

            if config_path.exists():
                config_path.rename(self.__global_config_path / "config.toml")
            if compilers_path.exists():
                compilers_path.rename(self.__global_data_path / "compilers")
            if solc_versions_path.exists():
                solc_versions_path.rename(
                    self.__global_data_path / ".woke_solc_version"
                )

            try:
                old_path.rmdir()
            except OSError:
                pass

        if project_root_path is None:
            self.__project_root_path = Path.cwd().resolve()
        else:
            self.__project_root_path = Path(project_root_path).resolve()

        if not self.__project_root_path.is_dir():
            raise ValueError(
                f"Project root path '{self.__project_root_path}' is not a directory."
            )

        self.__loaded_files = set()
        with change_cwd(self.__project_root_path):
            self.__config = TopLevelConfig()
        self.__config_raw = self.__config.dict(by_alias=True)

    def __str__(self) -> str:
        return self.__config.json(by_alias=True, exclude_unset=True)

    def __repr__(self) -> str:
        config_dict = reprlib.repr(self.__config_raw)
        return f"{self.__class__.__name__}.fromdict({config_dict}, project_root_path={repr(self.__project_root_path)})"

    def __merge_dicts(self, old: Dict[str, Any], new: Dict[str, Any]) -> None:
        for k, v in new.items():
            if k not in old.keys():
                old[k] = v
            else:
                if isinstance(v, dict):
                    self.__merge_dicts(old[k], new[k])
                else:
                    old[k] = v

    def __modified_keys(self, old: Dict, new: Dict, result: Dict) -> None:
        for k, v in new.items():
            if k not in old.keys():
                result[k] = v
            else:
                if isinstance(v, dict):
                    result[k] = {}
                    self.__modified_keys(old[k], new[k], result[k])
                    if not result[k]:
                        del result[k]
                else:
                    if old[k] != v:
                        result[k] = v
        for k, v in old.items():
            if k not in new.keys():
                result[k] = None

    def __load_file(
        self,
        parent: Optional[Path],
        path: Path,
        new_config: Dict[str, Any],
        graph: nx.DiGraph,
    ) -> None:
        if not path.is_file():
            if parent is None:
                logger.info(f"Config file '{path}' does not exist.")
            else:
                logger.warning(
                    f"Config file '{path}' loaded from '{parent}' does not exist."
                )
        else:
            # change the current working dir so that we can resolve relative paths
            with change_cwd(path.parent):
                with path.open("rb") as f:
                    loaded_config = tomli.load(f)

                graph.add_node(path, config=loaded_config)
                if parent is not None:
                    graph.add_edge(parent, path)

                # detect cyclic subconfigs
                if not nx.is_directed_acyclic_graph(graph):
                    cycles = list(nx.simple_cycles(graph))
                    error = f"Found cyclic config subconfigs:"
                    for no, cycle in enumerate(cycles):
                        error += f"\nCycle {no}:\n"
                        for path in cycle:
                            error += f"{path}\n"
                    raise ValueError(error)

                # validate the loaded config
                parsed_config = TopLevelConfig.parse_obj(loaded_config)

                # rebuild the loaded config from the pydantic model
                # this ensures that all stored paths are absolute
                loaded_config = parsed_config.dict(by_alias=True, exclude_unset=True)

                # merge the original config and the newly loaded config
                self.__merge_dicts(new_config, loaded_config)

                for subconfig_path in parsed_config.subconfigs:
                    self.__load_file(path, subconfig_path, new_config, graph)

    @classmethod
    def fromdict(
        cls,
        config_dict: Dict[str, Any],
        *_,
        project_root_path: Optional[Union[str, Path]] = None,
    ) -> "WokeConfig":
        """
        Build `WokeConfig` class from a dictionary and optional Woke root and project root paths.
        """
        instance = cls(project_root_path=project_root_path)
        with change_cwd(instance.project_root_path):
            parsed_config = TopLevelConfig.parse_obj(config_dict)
        instance.__config_raw = parsed_config.dict(by_alias=True, exclude_unset=True)
        instance.__config = parsed_config
        return instance

    def update(
        self,
        config_dict: Dict[str, Any],
        deleted_options: Iterable[Tuple[Union[int, str], ...]],
    ) -> Dict:
        """
        Update the config with a new dictionary. Return `True` if the config was changed.
        """
        with change_cwd(self.project_root_path):
            parsed_config = TopLevelConfig.parse_obj(config_dict)
        parsed_config_raw = parsed_config.dict(by_alias=True, exclude_unset=True)

        original_config = deepcopy(self.__config_raw)
        self.__merge_dicts(self.__config_raw, parsed_config_raw)

        for deleted_option in deleted_options:
            conf = self.__config_raw
            skip = False
            for segment in deleted_option[:-1]:
                if segment in conf:
                    conf = conf[segment]  # type: ignore
                else:
                    skip = True
                    break

            if skip:
                continue
            if isinstance(conf, dict):
                conf.pop(deleted_option[-1], None)  # type: ignore
            elif isinstance(conf, list):
                try:
                    conf.remove(deleted_option[-1])
                except ValueError:
                    pass

        self.__config = TopLevelConfig.parse_obj(self.__config_raw)
        modified_keys = {}
        self.__modified_keys(
            original_config,
            self.__config.dict(by_alias=True, exclude_unset=True),
            modified_keys,
        )
        return modified_keys

    def load_configs(self) -> None:
        """
        Clear any previous config options and load both the global config file `config.toml`
        located in the Woke root directory and the project specific config file `woke.toml`
        located in the project root directory. This is expected to be called right after `WokeConfig`
        instantiation.
        """
        self.__loaded_files = set()
        with change_cwd(self.__project_root_path):
            self.__config = TopLevelConfig()
        self.__config_raw = self.__config.dict(by_alias=True)

        self.load(self.global_config_path / "config.toml")
        self.load(self.project_root_path / "woke.toml")

    def load(self, path: Path) -> None:
        """
        Load config from the provided file path. Any already loaded config options are overridden by the options loaded
        from this file.
        """
        subconfigs_graph = nx.DiGraph()
        config_raw_copy = deepcopy(self.__config_raw)

        self.__load_file(None, path.resolve(), config_raw_copy, subconfigs_graph)

        config = TopLevelConfig.parse_obj(config_raw_copy)
        self.__config_raw = config_raw_copy
        self.__config = config
        self.__loaded_files.update(
            subconfigs_graph.nodes  # pyright: ignore reportGeneralTypeIssues
        )

    @property
    def loaded_files(self) -> FrozenSet[Path]:
        """
        Return frozenset of all loaded config file paths, including files that were loaded using the `subconfigs` config key.
        """
        return frozenset(self.__loaded_files)

    @property
    def global_config_path(self) -> Path:
        """
        Return the system path to the global config file.
        """
        return self.__global_config_path

    @property
    def global_data_path(self) -> Path:
        """
        Return the system path to the global data directory.
        """
        return self.__global_data_path

    @property
    def project_root_path(self) -> Path:
        """
        Return the system path of the currently open project.
        """
        return self.__project_root_path

    @property
    def min_solidity_version(self) -> SolidityVersion:
        return SolidityVersion.fromstring("0.6.2")

    @property
    def max_solidity_version(self) -> SolidityVersion:
        return SolidityVersion.fromstring("0.8.20")

    @property
    def detectors(self) -> DetectorsConfig:
        return self.__config.detectors

    @property
    def api_keys(self) -> Dict[str, str]:
        return self.__config.api_keys

    @property
    def compiler(self) -> CompilerConfig:
        """
        Return compiler-specific config options.
        """
        return self.__config.compiler

    @property
    def generator(self) -> GeneratorConfig:
        """
        Return config options specific to assets generated by Woke.
        """
        return self.__config.generator

    @property
    def lsp(self) -> LspConfig:
        """
        Return LSP-specific config options.
        """
        return self.__config.lsp

    @property
    def testing(self) -> TestingConfig:
        """
        Return testing framework-specific config options.
        """
        return self.__config.testing

    @property
    def deployment(self) -> DeploymentConfig:
        """
        Return deployment-specific config options.
        """
        return self.__config.deployment
